# sage_setup: distribution = sagemath-modules
# sage.doctest: needs sage.modules sage.rings.finite_rings
r"""
PRESENT

An ultra-lightweight block cipher.

This file implements the PRESENT block cipher and the corresponding key
schedule as described in [BKLPPRSV2007]_. PRESENT is an example of an
SP-network and consists of 31 rounds. The block length is 64 bits and two key
lengths of 80 and 128 bits are supported.

This implementation is meant for experimental and educational usage only,
do not use it in production code!

EXAMPLES:

Encrypt a message::

    sage: from sage.crypto.block_cipher.present import PRESENT
    sage: present = PRESENT()
    sage: present.encrypt(plaintext=0, key=0).hex()
    '2844b365c06992a3'

And decrypt it again::

    sage: present.decrypt(ciphertext=0x2844b365c06992a3, key=0)
    0

Have a look at the used round keys::

    sage: from sage.crypto.block_cipher.present import PRESENT_KS
    sage: ks = PRESENT_KS()
    sage: [k.hex() for k in ks(0)]
    ['0',
     'c000000000000000',
      ...
     '6dab31744f41d700']

Tweak around with the cipher::

    sage: from sage.crypto.sbox import SBox
    sage: cipher = PRESENT(rounds=1, doFinalRound=False)
    sage: cipher.sbox = SBox(range(16))
    sage: cipher.keySchedule = lambda x: [0, 0]  # return the 0 keys as round keys
    sage: cipher.encrypt(plaintext=0x1234, key=0x0).hex()
    '1234'

AUTHORS:

- Lukas Stennes (2019-02-01): initial version
"""

# ****************************************************************************
#       Copyright (C) 2013 Lukas Stennes <lukas.stennes@rub.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from sage.structure.sage_object import SageObject
from sage.rings.integer_ring import ZZ
from sage.rings.integer import Integer
from sage.modules.free_module_element import vector
from sage.rings.finite_rings.finite_field_constructor import GF
from sage.structure.element import Vector
from sage.crypto.sboxes import PRESENT as PRESENTSBOX


def _smallscale_present_linearlayer(nsboxes=16):
    """
    .. TODO::

        switch to sage.crypto.linearlayer
        (:issue:`25735`) as soon as it is included in sage

    EXAMPLES::

        sage: from sage.crypto.block_cipher.present import _smallscale_present_linearlayer
        sage: _smallscale_present_linearlayer(4)
        [1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0]
        [0 0 0 0 1 0 0 0 0 0 0 0 0 0 0 0]
        [0 0 0 0 0 0 0 0 1 0 0 0 0 0 0 0]
        [0 0 0 0 0 0 0 0 0 0 0 0 1 0 0 0]
        [0 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0]
        [0 0 0 0 0 1 0 0 0 0 0 0 0 0 0 0]
        [0 0 0 0 0 0 0 0 0 1 0 0 0 0 0 0]
        [0 0 0 0 0 0 0 0 0 0 0 0 0 1 0 0]
        [0 0 1 0 0 0 0 0 0 0 0 0 0 0 0 0]
        [0 0 0 0 0 0 1 0 0 0 0 0 0 0 0 0]
        [0 0 0 0 0 0 0 0 0 0 1 0 0 0 0 0]
        [0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 0]
        [0 0 0 1 0 0 0 0 0 0 0 0 0 0 0 0]
        [0 0 0 0 0 0 0 1 0 0 0 0 0 0 0 0]
        [0 0 0 0 0 0 0 0 0 0 0 1 0 0 0 0]
        [0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1]
    """
    from sage.modules.free_module import VectorSpace
    from sage.modules.free_module_element import vector
    from sage.matrix.constructor import Matrix
    from sage.rings.finite_rings.finite_field_constructor import GF

    def present_llayer(n, x):
        dim = 4*n
        y = [0]*dim
        for i in range(dim-1):
            y[i] = x[(n * i) % (dim - 1)]
        y[dim-1] = x[dim-1]
        return vector(GF(2), y)

    m = Matrix(GF(2), [present_llayer(nsboxes, ei)
                       for ei in VectorSpace(GF(2), 4*nsboxes).basis()])
    return m


class PRESENT(SageObject):
    r"""
    This class implements PRESENT described in [BKLPPRSV2007]_.

    EXAMPLES:

    You can invoke PRESENT encryption/decryption either by calling PRESENT with
    an appropriate flag::

        sage: from sage.crypto.block_cipher.present import PRESENT
        sage: present = PRESENT()
        sage: P = 0xFFFFFFFFFFFFFFFF
        sage: K = 0x0
        sage: present(present(P, K, 'encrypt'), K, 'decrypt') == P
        True

    Or by calling encryption/decryption methods directly::

        sage: C = present.encrypt(P, K)
        sage: P == present.decrypt(C, K)
        True

    The number of rounds can be reduced easily::

        sage: present = PRESENT(rounds=15)
        sage: present(present(P, K, 'encrypt'), K, 'decrypt') == P
        True

    You can use integers or a list-like bit representation for the inputs. If
    the input is an integer the output will be too. If it is list-like the
    output will be a bit vector::

        sage: P = ZZ(0).digits(2,padto=64)
        sage: K = ZZ(0).digits(2,padto=80)
        sage: list(present(present(P, K, 'encrypt'), K, 'decrypt')) == P
        True
        sage: P = ZZ(0).digits(2,padto=64)
        sage: K = 0x0
        sage: list(present(present(P, K, 'encrypt'), K, 'decrypt')) == P
        True

    The 80-bit version of PRESENT is used by default but the 128-bit version is
    also implemented::

        sage: present = PRESENT(128)
        sage: P = 0x0123456789abcdef
        sage: K = 0x00112233445566778899aabbccddeeff
        sage: present(present(P, K, 'encrypt'), K, 'decrypt') == P
        True

    .. SEEALSO::

        :class:`PRESENT_KS`
        :mod:`sage.crypto.sboxes`

    .. automethod:: __init__
    .. automethod:: __call__
    """

    def __init__(self, keySchedule=80, rounds=None, doFinalRound=False):
        r"""
        Construct an instance of PRESENT.

        INPUT:

        - ``keySchedule`` -- (default: ``80``) the key schedule that will be
          used for encryption and decryption. Use ``80`` or ``128`` as a
          shortcut for the original key schedules from [BKLPPRSV2007]_.

        - ``rounds`` -- integer (default: ``None``); the number of rounds. If
          ``None`` the number of rounds of the key schedule is used.

        - ``doFinalRound`` -- boolean (default: ``False``); flag to
          control whether the linear layer in the last round should take place
          or not. Since the last linear layer does not add any security, it
          usually does not take place in real world implementations for
          performance reasons.

        EXAMPLES:

        By default a 80-bit version with 31 rounds is created::

            sage: from sage.crypto.block_cipher.present import PRESENT
            sage: PRESENT() # indirect doctest
            PRESENT block cipher with 31 rounds, deactivated linear layer in
            last round and the following key schedule:
            Original PRESENT key schedule with 80-bit keys and 31 rounds

        The 128-bit version is also implemented::

            sage: PRESENT(128) # indirect doctest
            PRESENT block cipher with 31 rounds, deactivated linear layer in
            last round and the following key schedule:
            Original PRESENT key schedule with 128-bit keys and 31 rounds

        Reducing the number of rounds is simple. But increasing it is not
        possible::

            sage: PRESENT(keySchedule=80, rounds=23) # indirect doctest
            PRESENT block cipher with 23 rounds, deactivated linear layer in
            last round and the following key schedule:
            Original PRESENT key schedule with 80-bit keys and 31 rounds
            sage: PRESENT(80, 32) # indirect doctest
            Traceback (most recent call last):
            ...
            ValueError: number of rounds must be less or equal to the number
            of rounds of the key schedule


        By default the linear layer operation in the last round is omitted but
        of course you can enable it::

            sage: PRESENT(doFinalRound=True) # indirect doctest
            PRESENT block cipher with 31 rounds, activated linear layer in
            last round and the following key schedule:
            Original PRESENT key schedule with 80-bit keys and 31 rounds

        You can use arbitrary key schedules. Since it is the only one
        implemented here the original key schedule is used for demonstration::

            sage: from sage.crypto.block_cipher.present import PRESENT_KS
            sage: PRESENT(keySchedule=PRESENT_KS(80, 15)) # indirect doctest
            PRESENT block cipher with 15 rounds, deactivated linear layer in
            last round and the following key schedule:
            Original PRESENT key schedule with 80-bit keys and 15 rounds

        .. SEEALSO::

            :class:`PRESENT_KS`
        """
        if keySchedule == 80:
            self.keySchedule = PRESENT_KS()
            self._keysize = 80
        elif keySchedule == 128:
            self.keySchedule = PRESENT_KS(128)
            self._keysize = 128
        else:
            self.keySchedule = keySchedule
        if rounds is None:
            self._rounds = self.keySchedule._rounds
        elif rounds <= self.keySchedule._rounds:
            self._rounds = rounds
        else:
            raise ValueError('number of rounds must be less or equal to the '
                             'number of rounds of the key schedule')
        self._blocksize = 64
        self.sbox = PRESENTSBOX
        self._permutationMatrix = _smallscale_present_linearlayer()
        self._inversePermutationMatrix = self._permutationMatrix.inverse()
        self._doFinalRound = doFinalRound

    def __call__(self, block, key, algorithm='encrypt'):
        r"""
        Apply PRESENT encryption or decryption on ``block`` using ``key``.
        The flag ``algorithm`` controls what action is to be performed on
        ``block``.

        INPUT:

        - ``block`` -- integer or bit list-like; the plaintext or ciphertext

        - ``K`` -- integer or bit list-like; the key

        - ``algorithm`` -- string (default: ``'encrypt'``); a flag to signify
          whether encryption or decryption is to be applied to ``B``. The
          encryption flag is ``'encrypt'`` and the decryption flag is
          ``'decrypt'``

        OUTPUT:

        - The plaintext or ciphertext corresponding to ``block``, obtained using
          the ``key``. If ``block`` is an integer the output will be too. If
          ``block`` is list-like the output will be a bit vector.

        EXAMPLES::

            sage: from sage.crypto.block_cipher.present import PRESENT
            sage: present = PRESENT(doFinalRound=True)
            sage: P = 0xFFFFFFFFFFFFFFFF
            sage: K = 0x0
            sage: present(P, K, 'encrypt').hex()
            'a112ffc72f68417b'
        """
        if algorithm == 'encrypt':
            return self.encrypt(block, key)
        elif algorithm == 'decrypt':
            return self.decrypt(block, key)
        else:
            raise ValueError('Algorithm must be \'encrypt\' or \'decrypt\' and'
                             ' not \'%s\'' % algorithm)

    def __eq__(self, other):
        r"""
        Compare ``self`` with ``other``.

        PRESENT objects are the same if all attributes are the same.

        EXAMPLES::

            sage: from sage.crypto.block_cipher.present import PRESENT
            sage: PRESENT(80) == PRESENT(80) # indirect doctest
            True
            sage: PRESENT(80) == PRESENT(128) # indirect doctest
            False
            sage: PRESENT(80) == 80 # indirect doctest
            False
            sage: present = PRESENT()
            sage: present.sbox = present.sbox.inverse()
            sage: present == PRESENT() # indirect doctest
            False
        """
        if not isinstance(other, PRESENT):
            return False
        else:
            return self.__dict__ == other.__dict__

    def __repr__(self):
        r"""
        A string representation of this PRESENT.

        EXAMPLES::

            sage: from sage.crypto.block_cipher.present import PRESENT
            sage: PRESENT() # indirect doctest
            PRESENT block cipher with 31 rounds, deactivated linear layer in
            last round and the following key schedule:
            Original PRESENT key schedule with 80-bit keys and 31 rounds
        """
        return ('PRESENT block cipher with %s rounds, %s linear layer in last '
                'round and the following key schedule:\n%s'
                % (self._rounds, 'activated' if self._doFinalRound else
                   'deactivated', self.keySchedule.__repr__()))

    def encrypt(self, plaintext, key):
        r"""
        Return the ciphertext corresponding to ``plaintext``, using PRESENT
        encryption with ``key``.

        INPUT:

        - ``plaintext`` -- integer or bit list-like; the plaintext that will be
          encrypted

        - ``key`` -- integer or bit list-like; the key

        OUTPUT:

        - The ciphertext corresponding to ``plaintext``, obtained using the
          ``key``. If ``plaintext`` is an integer the output will be too. If
          ``plaintext`` is list-like the output will be a bit vector.

        EXAMPLES:

        The test vectors from [BKLPPRSV2007]_ are checked here::

            sage: from sage.crypto.block_cipher.present import PRESENT
            sage: present = PRESENT(doFinalRound=True)
            sage: p1 = 0x0
            sage: k1 = 0x0
            sage: c1 = 0x5579C1387B228445
            sage: present.encrypt(p1, k1) == c1
            True
            sage: p2 = 0x0
            sage: k2 = 0xFFFFFFFFFFFFFFFFFFFF
            sage: c2 = 0xE72C46C0F5945049
            sage: present.encrypt(p2, k2) == c2
            True
            sage: p3 = 0xFFFFFFFFFFFFFFFF
            sage: k3 = 0x0
            sage: c3 = 0xA112FFC72F68417B
            sage: present.encrypt(p3, k3) == c3
            True
            sage: p4 = 0xFFFFFFFFFFFFFFFF
            sage: k4 = 0xFFFFFFFFFFFFFFFFFFFF
            sage: c4 = 0x3333DCD3213210D2
            sage: present.encrypt(p4, k4) == c4
            True

        ALGORITHM:

        Description of the encryption function based on [BKLPPRSV2007]_:

        A top-level algorithmic description of PRESENT encryption::

            generateRoundKeys()
            for i = 1 to 31 do
                addRoundkey(STATE, K_i)
                sBoxLayer(STATE)
                pLayer(STATE)
            end for
            addRoundkey(STATE, K_{32})

        Each of the 31 rounds consists of an XOR operation to introduce a round
        key `K_i` for `1 \leq i \leq 32`, where `K_{32}` is used for
        post-whitening, a linear bitwise permutation and a non-linear
        substitution layer. The non-linear layer uses a single 4-bit S-box
        which is applied 16 times in parallel in each round. Each stage but
        addRoundkey is specified in its corresponding function.

        **addRoundkey:**
        Given round key `K_i = \kappa^i_{63} \dots \kappa^i_0` for `1 \leq i
        \leq 32` and current STATE `b_{63} \dots b_0`, addRoundkey consists of
        the operation for `0 \leq j \leq 63`, `b_j = b_j \oplus \kappa^i_j`.
        """
        if isinstance(plaintext, (list, tuple, Vector)):
            inputType = 'vector'
        elif isinstance(plaintext, (Integer, int)):
            inputType = 'integer'
        state = convert_to_vector(plaintext, 64)
        key = convert_to_vector(key, self._keysize)
        roundKeys = self.keySchedule(key)
        for r, K in enumerate(roundKeys[:self._rounds]):
            state = self.round(state, r, K)
        state = state + roundKeys[self._rounds]
        return state if inputType == 'vector' else ZZ(list(state), 2)

    def decrypt(self, ciphertext, key):
        r"""
        Return the plaintext corresponding to the ``ciphertext``, using PRESENT decryption with ``key``.

        INPUT:

        - ``ciphertext`` -- integer or bit list-like; the ciphertext that will
          be decrypted

        - ``key`` -- integer or bit list-like; the key

        OUTPUT:

        - The plaintext corresponding to ``ciphertext``, obtained using the
          ``key``. If ``ciphertext`` is an integer the output will be too. If
          ``ciphertext`` is list-like the output will be a bit vector.

        EXAMPLES:

        The test vectors from [BKLPPRSV2007]_ are checked here::

            sage: from sage.crypto.block_cipher.present import PRESENT
            sage: present = PRESENT(doFinalRound=True)
            sage: p1 = 0x0
            sage: k1 = 0x0
            sage: c1 = 0x5579C1387B228445
            sage: present.decrypt(c1, k1) == p1
            True
            sage: p2 = 0x0
            sage: k2 = 0xFFFFFFFFFFFFFFFFFFFF
            sage: c2 = 0xE72C46C0F5945049
            sage: present.decrypt(c2, k2) == p2
            True
            sage: p3 = 0xFFFFFFFFFFFFFFFF
            sage: k3 = 0x0
            sage: c3 = 0xA112FFC72F68417B
            sage: present.decrypt(c3, k3) == p3
            True
            sage: p4 = 0xFFFFFFFFFFFFFFFF
            sage: k4 = 0xFFFFFFFFFFFFFFFFFFFF
            sage: c4 = 0x3333DCD3213210D2
            sage: present.decrypt(c4, k4) == p4
            True
        """
        if isinstance(ciphertext, (list, tuple, Vector)):
            inputType = 'vector'
        elif isinstance(ciphertext, (Integer, int)):
            inputType = 'integer'
        state = convert_to_vector(ciphertext, 64)
        key = convert_to_vector(key, self._keysize)
        roundKeys = self.keySchedule(key)
        state = state + roundKeys[self._rounds]
        for r, K in enumerate(roundKeys[:self._rounds][::-1]):
            state = self.round(state, r, K, inverse=True)
        return state if inputType == 'vector' else ZZ(list(state), 2)

    def round(self, state, round_counter, round_key, inverse=False):
        """
        Apply one round of PRESENT to ``state`` and return the result.

        EXAMPLES::

            sage: from sage.crypto.block_cipher.present import PRESENT
            sage: from sage.crypto.block_cipher.present import convert_to_vector
            sage: present = PRESENT(128)
            sage: k = convert_to_vector(0x0011223344556677, 64)
            sage: p = convert_to_vector(0x0123456789abcdef, 64)
            sage: ZZ(list(present.round(p, 0, k)), 2).hex()
            'ad0ed4ca386b6559'
        """
        out = state[:]
        if not inverse:
            out = out + round_key
            out = self.sbox_layer(out)
            if self._doFinalRound or round_counter != self._rounds - 1:
                out = self.linear_layer(out)
        else:
            if self._doFinalRound or round_counter != 0:
                out = self.linear_layer(out, inverse=True)
            out = self.sbox_layer(out, inverse=True)
            out = out + round_key
        return out

    def sbox_layer(self, state, inverse=False):
        r"""
        Apply the sBoxLayer of PRESENT to the bit vector ``state`` and return
        the result.

        The S-box used in PRESENT is a 4-bit to 4-bit S-box. The action of this
        box in hexadecimal notation is given by the following table.

        +------+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
        | x    | 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9 | A | B | C | D | E | F |
        +------+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
        | S[x] | C | 5 | 6 | B | 9 | 0 | A | D | 3 | E | F | 8 | 4 | 7 | 1 | 2 |
        +------+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

        For sBoxLayer the current STATE `b_{63} \dots b_0` is considered as
        sixteen 4-bit words `w_{15} \dots w_0` where `w_i =
        b_{4i+3}||b_{4i+2}||b_{4i+1}||b_{4i}` for `0 \leq i \leq 15` and the
        output nibble S[`w_i`] provides the updated state values in the obvious
        way.

        EXAMPLES::

            sage: from sage.crypto.block_cipher.present import PRESENT
            sage: present = PRESENT()
            sage: state = vector(GF(2), 64, [0]*64)
            sage: present.sbox_layer(state)
            (0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0,
            1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1,
            0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1)
            sage: state = vector(GF(2), 64, [1]+[0]*63)
            sage: present.sbox_layer(state)
            (1, 0, 1, 0, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0,
            1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1,
            0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1)

        .. NOTE::

            :mod:`sage.crypto.sbox` uses big endian by default whereas most of
            Sage uses little endian. So to use the big endian PRESENT Sbox from
            :mod:`sage.crypto.sboxes` :func:`sbox_layer` has to do some endian
            conversion (i.e. reverse input and output of the Sbox). Keep this in
            mind if you change the Sbox or :func:`sbox_layer`.
        """
        sbox = self.sbox if not inverse else self.sbox.inverse()
        out = vector(GF(2), 64)
        for nibble in [slice(4*j, 4*j+4) for j in range(16)]:
            out[nibble] = sbox(state[nibble][::-1])[::-1]
        return out

    def linear_layer(self, state, inverse=False):
        """
        Apply the pLayer of PRESENT to the bit vector ``state`` and return the
        result.

        The bit permutation used in PRESENT is given by the following
        table. Bit `i` of STATE is moved to bit position `P(i)`.

        +------+----+----+----+----+----+----+----+----+----+----+----+----+----+----+----+----+
        | i    |  0 |  1 |  2 |  3 |  4 |  5 |  6 |  7 |  8 |  9 | 10 | 11 | 12 | 13 | 14 | 15 |
        +------+----+----+----+----+----+----+----+----+----+----+----+----+----+----+----+----+
        | P(i) |  0 | 16 | 32 | 48 |  1 | 17 | 33 | 49 |  2 | 18 | 34 | 50 |  3 | 19 | 35 | 51 |
        +------+----+----+----+----+----+----+----+----+----+----+----+----+----+----+----+----+
        +------+----+----+----+----+----+----+----+----+----+----+----+----+----+----+----+----+
        | i    | 16 | 17 | 18 | 19 | 20 | 21 | 22 | 23 | 24 | 25 | 26 | 27 | 28 | 29 | 30 | 31 |
        +------+----+----+----+----+----+----+----+----+----+----+----+----+----+----+----+----+
        | P(i) |  4 | 20 | 36 | 52 |  5 | 21 | 37 | 53 |  6 | 22 | 38 | 54 |  7 | 23 | 39 | 55 |
        +------+----+----+----+----+----+----+----+----+----+----+----+----+----+----+----+----+
        +------+----+----+----+----+----+----+----+----+----+----+----+----+----+----+----+----+
        | i    | 32 | 33 | 34 | 35 | 36 | 37 | 38 | 39 | 40 | 41 | 42 | 43 | 44 | 45 | 46 | 47 |
        +------+----+----+----+----+----+----+----+----+----+----+----+----+----+----+----+----+
        | P(i) |  8 | 24 | 40 | 56 |  9 | 25 | 41 | 57 | 10 | 26 | 42 | 58 | 11 | 27 | 43 | 59 |
        +------+----+----+----+----+----+----+----+----+----+----+----+----+----+----+----+----+
        +------+----+----+----+----+----+----+----+----+----+----+----+----+----+----+----+----+
        | i    | 48 | 49 | 50 | 51 | 52 | 53 | 54 | 55 | 56 | 57 | 58 | 59 | 60 | 61 | 62 | 63 |
        +------+----+----+----+----+----+----+----+----+----+----+----+----+----+----+----+----+
        | P(i) | 12 | 28 | 44 | 60 | 13 | 29 | 45 | 61 | 14 | 30 | 46 | 62 | 15 | 31 | 47 | 63 |
        +------+----+----+----+----+----+----+----+----+----+----+----+----+----+----+----+----+

        EXAMPLES::

            sage: from sage.crypto.block_cipher.present import PRESENT
            sage: present = PRESENT()
            sage: state = vector(GF(2), 64, [0, 1]+[0]*62)
            sage: present.linear_layer(state)
            (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)
        """
        m = self._permutationMatrix if not inverse else self._inversePermutationMatrix
        return m * state


class PRESENT_KS(SageObject):
    r"""
    This class implements the PRESENT key schedules for both 80-bit and 128-bit
    keys as described in [BKLPPRSV2007]_.

    EXAMPLES:

    Initialise the key schedule with a `master\_key` to use it as an iterable::

        sage: from sage.crypto.block_cipher.present import PRESENT_KS
        sage: ks = PRESENT_KS(master_key=0)
        sage: ks[0] == 0x0
        True
        sage: ks[31] == 0x6dab31744f41d700
        True

    Or omit the `master\_key` and pass a key when calling the key schedule::

        sage: ks = PRESENT_KS(keysize=128)
        sage: K = ks(0x00112233445566778899aabbccddeeff)
        sage: K[0] == 0x0011223344556677
        True
        sage: K[31] == 0x091989a5ae8eab21
        True

    ALGORITHM:

    Description of the key schedule for 64-bit and 128-bit keys from
    [BKLPPRSV2007]_:

    The key schedule for 64-bit keys works as follows:

    At round `i` the 64-bit round key `K_i = \kappa_{63}\kappa_{62} \dots
    \kappa_0` consists of the 64 leftmost bits of the current contents of
    register `K`. Thus at round `i` we have that:

    .. MATH::

        K_i = \kappa_{63}\kappa_{62}\dots \kappa_0 = k_{79}k_{78}\dots k_{16}.

    After extracting the round key `K_i`, the key register `K = k_{79}k_{78}
    \dots k_0` is updated as follows:

    .. MATH::

        \begin{aligned}
        \ [k_{79}k_{78}\dots k_{1}k_{0}] &= [k_{18}k_{17}\dots k_{20}k_{19}] \\
        [k_{79}k_{78}k_{77}k_{76}] &= S[k_{79}k_{78}k_{77}k_{76}] \\
        [k_{19}k_{18}k_{17}k_{16}k_{15}] &= [k_{19}k_{18}k_{17}k_{16}k_{15}]
                                            \oplus round\_counter
        \end{aligned}

    Thus, the key register is rotated by 61 bit positions to the left, the
    left-most four bits are passed through the PRESENT S-box, and the
    round_counter value `i` is exclusive-ored with bits `k_{19}k_{18}k_{17}
    k_{16}k_{15}` of `K` with the least significant bit of round_counter on
    the right.

    The key schedule for 128-bit keys works as follows:

    At round `i` the 64-bit round key `K_i = \kappa_{63}\kappa_{62} \dots
    \kappa_0` consists of the 64 leftmost bits fo the current contents of
    register `K`. Thus at round `i` we have that:

    .. MATH::

        K_i = \kappa_{63}\kappa_{62}\dots \kappa_0 = k_{127}k_{126}\dots k_{64}

    After extracting the round key `K_i`, the key register `K = k_{127}k_{126}
    \dots k_0` is updated as follows:

    .. MATH::

        \begin{aligned}
        \ [k_{127}k_{126}\dots k_{1}k_{0}] &= [k_{66}k_{65}\dots k_{68}k_{67}]
        \\
        [k_{127}k_{126}k_{125}k_{124}] &= S[k_{127}k_{126}k_{125}k_{124}]\\
        [k_{123}k_{122}k_{121}k_{120}] &= S[k_{123}k_{122}k_{121}k_{120}]\\
        [k_{66}k_{65}k_{64}k_{63}k_{62}] &= [k_{66}k_{65}k_{64}k_{63}k_{62}]
                                            \oplus round\_counter
        \end{aligned}

    Thus, the key register is rotated by 61 bit positions to the left, the
    left-most eight bits are passed through two PRESENT S-boxes, and the
    round_counter value `i` is exclusive-ored with bits `k_{66}k_{65}k_{64}
    k_{63}k_{62}` of `K` with the least significant bit of round_counter on
    the right.

    .. SEEALSO::

        :class:`PRESENT`
        :mod:`sage.crypto.sboxes`

    .. NOTE::

        :mod:`sage.crypto.sbox` uses big endian by default whereas most of Sage
        uses little endian. So to use the big endian PRESENT Sbox from
        :mod:`sage.crypto.sboxes` :class:`PRESENT_KS` has to do some endian
        conversion (i.e. reverse input and output of the Sbox). Keep this in
        mind if you change the Sbox or :func:`__call__`.

    .. automethod:: __init__
    .. automethod:: __call__
    """

    def __init__(self, keysize=80, rounds=31, master_key=None):
        r"""
        Construct an instance of PRESENT_KS.

        INPUT:

        - ``keysize`` -- integer (default: 80); the size of the keys that
          will be used in bits. It must be either 80 or 128.

        - ``rounds`` -- integer (default: 31); the number of rounds
          ``self`` can create keys for

        - ``master_key`` -- integer or bit list-like (default: ``None``); the
          key that will be used

        EXAMPLES::

            sage: from sage.crypto.block_cipher.present import PRESENT_KS
            sage: PRESENT_KS()
            Original PRESENT key schedule with 80-bit keys and 31 rounds

        .. NOTE::

            If you want to use a PRESENT_KS object as an iterable you have to
            pass a ``master_key`` value on initialisation. Otherwise you can
            omit ``master_key`` and pass a key when you call the object.
        """
        if keysize != 80 and keysize != 128:
            raise ValueError('keysize must bei either 80 or 128 and not %s'
                             % keysize)
        self._keysize = keysize
        self._rounds = rounds
        self.sbox = PRESENTSBOX
        self._master_key = master_key

    def __call__(self, K):
        r"""
        Return all round keys in a list.

        INPUT:

        - ``K`` -- integer or bit list-like; the key

        OUTPUT:

        - A list containing ``rounds + 1`` round keys. Since addRoundkey takes
          place in every round and after the last round there must be
          ``rounds + 1`` round keys.  If ``K`` is an integer the elements of
          the output list will be too. If ``K`` is list-like the element of the
          output list will be  bit vectors.

        EXAMPLES::

            sage: from sage.crypto.block_cipher.present import PRESENT_KS
            sage: ks = PRESENT_KS()
            sage: ks(0)[31] == 0x6dab31744f41d700
            True

        .. NOTE::

            If you want to use a PRESENT_KS object as an iterable you have to
            pass a ``master_key`` value on initialisation. Otherwise you can
            omit ``master_key`` and pass a key when you call the object.
        """
        if isinstance(K, (list, tuple, Vector)):
            inputType = 'vector'
        elif isinstance(K, (Integer, int)):
            inputType = 'integer'
        K = convert_to_vector(K, self._keysize)
        roundKeys = []
        if self._keysize == 80:
            for i in range(1, self._rounds+1):
                roundKeys.append(K[16:])
                K[0:] = list(K[19:]) + list(K[:19])
                K[76:] = self.sbox(K[76:][::-1])[::-1]
                rc = vector(GF(2), ZZ(i).digits(2, padto=5))
                K[15:20] = K[15:20] + rc
            roundKeys.append(K[16:])
        elif self._keysize == 128:
            for i in range(1, self._rounds+1):
                roundKeys.append(K[64:])
                K[0:] = list(K[67:]) + list(K[:67])
                K[124:] = self.sbox(K[124:][::-1])[::-1]
                K[120:124] = self.sbox(K[120:124][::-1])[::-1]
                rc = vector(GF(2), ZZ(i).digits(2, padto=5))
                K[62:67] = K[62:67] + rc
            roundKeys.append(K[64:])
        return roundKeys if inputType == 'vector' else [ZZ(list(k), 2) for k in
                                                        roundKeys]

    def __eq__(self, other):
        r"""
        Compare ``self`` with ``other``.

        PRESENT_KS objects are the same if all attributes are the same.

        EXAMPLES::

            sage: from sage.crypto.block_cipher.present import PRESENT_KS
            sage: PRESENT_KS(80) == PRESENT_KS(80) # indirect doctest
            True
            sage: PRESENT_KS(80) == PRESENT_KS(128) # indirect doctest
            False
            sage: PRESENT_KS(80) == 80 # indirect doctest
            False
        """
        if not isinstance(other, PRESENT_KS):
            return False
        else:
            return self.__dict__ == other.__dict__

    def __repr__(self):
        r"""
        A string representation of this PRESENT_KS.

        EXAMPLES::

            sage: from sage.crypto.block_cipher.present import PRESENT_KS
            sage: PRESENT_KS() # indirect doctest
            Original PRESENT key schedule with 80-bit keys and 31 rounds
        """
        return ('Original PRESENT key schedule with %s-bit keys and %s rounds'
                % (self._keysize, self._rounds))

    def __getitem__(self, r):
        r"""
        Compute the sub key for round ``r`` derived from initial master key.

        The key schedule object has to have been initialised with the
        ``master_key`` argument.

        INPUT:

        - ``r`` -- integer; the round for which the sub key is computed

        EXAMPLES::

            sage: from sage.crypto.block_cipher.present import PRESENT_KS
            sage: ks = PRESENT_KS(master_key=0x0)
            sage: ks[0] ==  0x0 # indirect doctest
            True
            sage: ks[31] ==  0x6dab31744f41d700 # indirect doctest
            True
        """
        if self._master_key is None:
            raise ValueError('Key not set during initialisation')
        return self(self._master_key)[r]

    def __iter__(self):
        """
        Iterate over the ``self._rounds + 1`` PRESENT round keys, derived from
        ``master_key``.

        EXAMPLES::

            sage: from sage.crypto.block_cipher.present import PRESENT_KS
            sage: K = [k for k in PRESENT_KS(master_key=0x0)]
            sage: K[0] == 0x0 # indirect doctest
            True
            sage: K[31] == 0x6dab31744f41d700 # indirect doctest
            True
        """
        if self._master_key is None:
            raise ValueError('Key not set during initialisation')
        return iter(self(self._master_key))


def convert_to_vector(I, L):
    r"""
    Convert ``I`` to a bit vector of length ``L``.

    INPUT:

    - ``I`` -- integer or bit list-like

    - ``L`` -- integer; the desired bit length of the output

    OUTPUT: the ``L``-bit vector representation of ``I``

    EXAMPLES::

        sage: from sage.crypto.block_cipher.present import convert_to_vector
        sage: convert_to_vector(0x1F, 8)
        (1, 1, 1, 1, 1, 0, 0, 0)
        sage: v = vector(GF(2), 4, [1,0,1,0])
        sage: convert_to_vector(v, 4)
        (1, 0, 1, 0)
    """
    try:
        state = vector(GF(2), L, ZZ(I).digits(2, padto=L))
        return state
    except TypeError:
        # ignore the error and try list-like types
        pass
    state = vector(GF(2), L, ZZ(list(I), 2).digits(2, padto=L))
    return state



# sage_setup: distribution = sagemath-modules
# sage.doctest: needs sage.graphs
r"""
Symplectic Derivation Lie Algebras

AUTHORS:

- Travis Scrimshaw (2020-10): Initial version
"""

# ****************************************************************************
#       Copyright (C) 2020 Travis Scrimshaw <tcscrims at gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.categories.lie_algebras import LieAlgebras
from sage.sets.family import Family
from sage.sets.disjoint_union_enumerated_sets import DisjointUnionEnumeratedSets
from sage.structure.indexed_generators import IndexedGenerators
from sage.algebras.lie_algebras.lie_algebra_element import LieAlgebraElement
from sage.algebras.lie_algebras.lie_algebra import InfinitelyGeneratedLieAlgebra
from sage.sets.non_negative_integers import NonNegativeIntegers
from sage.combinat.partition import _Partitions, Partitions


class SymplecticDerivationLieAlgebra(InfinitelyGeneratedLieAlgebra, IndexedGenerators):
    r"""
    The symplectic derivation Lie algebra.

    Fix a `g \geq 4` and let `R` be a commutative ring. Let `H = R^{2g}`
    be equipped with a symplectic form `\mu` with the basis
    `a_1, \ldots, a_g, b_1, \ldots, b_g` such that

    .. MATH::

        \mu(a_i, a_j) = \mu(b_i, b_j) = 0,
        \qquad\qquad
        \mu(a_i, b_j) = -\mu(b_j, a_i) = \delta_{ij},

    for all `i, j`. The *symplectic derivation Lie algebra* is the Lie
    algebra

    .. MATH::

        \mathfrak{c}_g := \bigoplus_{w \geq 0} S^{w+2} H

    with the Lie bracket on basis elements

    .. MATH::

        [x_1 \cdots x_{m+2}, y_1 \cdots y_{n+2}] =
        \sum_{i,j} \mu(x_i, y_j) x_1 \cdots \widehat{x}_i \cdots x_{m+2}
        \cdot y_1 \cdots \widehat{y}_j \cdots y_{n+2},

    where `\widehat{z}` denotes that factor is missing. When `R = \QQ`, this
    corresponds to the classical Poisson bracket on `C^{\infty}(\RR^{2g})`
    restricted to polynomials with coefficients in `\QQ`.

    EXAMPLES::

        sage: L = lie_algebras.SymplecticDerivation(QQ, 5)
        sage: elts = L.some_elements()
        sage: list(elts)
        [a1*a2, b1*b3, a1*a1*a2, b3*b4,
         a1*a4*b3, a1*a2 - 1/2*a1*a2*a2*a5 + a1*a1*a2*b1*b4]
        sage: [[elts[i].bracket(elts[j]) for i in range(len(elts))]
        ....:  for j in range(len(elts))]
        [[0, -a2*b3, 0, 0, 0, -a1*a1*a2*a2*b4],
         [a2*b3, 0, 2*a1*a2*b3, 0, a4*b3*b3, a2*b3 - 1/2*a2*a2*a5*b3 + 2*a1*a2*b1*b3*b4],
         [0, -2*a1*a2*b3, 0, 0, 0, -2*a1*a1*a1*a2*a2*b4],
         [0, 0, 0, 0, a1*b3*b3, 0],
         [0, -a4*b3*b3, 0, -a1*b3*b3, 0, -a1*a1*a1*a2*b1*b3 - a1*a1*a2*a4*b3*b4],
         [a1*a1*a2*a2*b4, -a2*b3 + 1/2*a2*a2*a5*b3 - 2*a1*a2*b1*b3*b4, 2*a1*a1*a1*a2*a2*b4,
          0, a1*a1*a1*a2*b1*b3 + a1*a1*a2*a4*b3*b4, 0]]
        sage: x = L.monomial(Partition([8,8,6,6,4,2,2,1,1,1])); x
        a1*a1*a1*a2*a2*a4*b1*b1*b3*b3
        sage: [L[x, elt] for elt in elts]
        [-2*a1*a1*a1*a2*a2*a2*a4*b1*b3*b3,
         3*a1*a1*a2*a2*a4*b1*b1*b3*b3*b3,
         -4*a1*a1*a1*a1*a2*a2*a2*a4*b1*b3*b3,
         a1*a1*a1*a2*a2*b1*b1*b3*b3*b3,
         -2*a1*a1*a1*a2*a2*a4*a4*b1*b3*b3*b3,
         -2*a1*a1*a1*a2*a2*a2*a4*b1*b3*b3 + a1*a1*a1*a2*a2*a2*a2*a4*a5*b1*b3*b3
          + a1*a1*a1*a1*a1*a2*a2*a2*b1*b1*b1*b3*b3 - a1*a1*a1*a1*a2*a2*a2*a4*b1*b1*b3*b3*b4]

    REFERENCES:

    - [Harako2020]_
    """
    def __init__(self, R, g):
        r"""
        Initialize ``self``.

        EXAMPLES::

            sage: L = lie_algebras.SymplecticDerivation(QQ, 5)
            sage: TestSuite(L).run()
        """
        if g < 4:
            raise ValueError("g must be at least 4")
        cat = LieAlgebras(R).WithBasis().Graded()
        self._g = g
        d = Family(NonNegativeIntegers(), lambda n: Partitions(n, min_length=2, max_part=2*g))
        indices = DisjointUnionEnumeratedSets(d)
        InfinitelyGeneratedLieAlgebra.__init__(self, R, index_set=indices, category=cat)
        IndexedGenerators.__init__(self, indices, sorting_key=self._basis_key)

    def _basis_key(self, x):
        r"""
        Return the key used to compare two basis element indices.

        EXAMPLES::

            sage: L = lie_algebras.SymplecticDerivation(QQ, 5)
            sage: L._basis_key( [7, 5, 2, 1] )
            (4, [7, 5, 2, 1])

            sage: x = L.an_element(); x
            a1*a2 - 1/2*a1*a2*a2*a5 + a1*a1*a2*b1*b4
            sage: sorted(map(L._basis_key, x.support()))
            [(2, [2, 1]), (4, [5, 2, 2, 1]), (5, [9, 6, 2, 1, 1])]
        """
        return (len(x), x)

    def _repr_term(self, m):
        r"""
        Return a string representation of the term indexed by ``m``.

        EXAMPLES::

            sage: L = lie_algebras.SymplecticDerivation(QQ, 5)
            sage: L._repr_term([7, 5, 2, 1])
            'a1*a2*a5*b2'
        """
        g = self._g

        def label(i):
            return "a{}".format(i) if i <= g else "b{}".format(i-g)
        return "*".join(label(i) for i in reversed(m))

    def _latex_term(self, m):
        r"""
        Return a `\LaTeX` representation of the term indexed by ``m``.

        EXAMPLES::

            sage: L = lie_algebras.SymplecticDerivation(QQ, 5)
            sage: L._latex_term([7, 5, 2, 1])
            'a_{1} a_{2} a_{5} b_{2}'
        """
        g = self._g

        def label(i):
            return "a_{{{}}}".format(i) if i <= g else "b_{{{}}}".format(i-g)
        return " ".join(label(i) for i in reversed(m))

    def _unicode_art_term(self, m):
        r"""
        Return a unicode art representation of the term indexed by ``m``.

        EXAMPLES::

            sage: L = lie_algebras.SymplecticDerivation(QQ, 5)
            sage: L._unicode_art_term([7, 5, 2, 1])
            a₁·a₂·a₅·b₂
        """
        from sage.typeset.unicode_art import unicode_art, unicode_subscript
        g = self._g

        def label(i):
            return "a{}".format(unicode_subscript(i)) if i <= g else "b{}".format(unicode_subscript(i-g))
        return unicode_art("·".join(label(i) for i in reversed(m)))

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: lie_algebras.SymplecticDerivation(QQ, 5)
            Symplectic derivation Lie algebra of rank 5 over Rational Field
        """
        return "Symplectic derivation Lie algebra of rank {} over {}".format(self._g, self.base_ring())

    def degree_on_basis(self, x):
        r"""
        Return the degree of the basis element indexed by ``x``.

        EXAMPLES::

            sage: L = lie_algebras.SymplecticDerivation(QQ, 5)
            sage: L.degree_on_basis([5,2,1])
            1
            sage: L.degree_on_basis([1,1])
            0
            sage: elt = L.monomial(Partition([5,5,2,1])) + 3*L.monomial(Partition([3,3,2,1]))
            sage: elt.degree()
            2
        """
        return len(x) - 2

    def bracket_on_basis(self, x, y):
        r"""
        Return the bracket of basis elements indexed by ``x`` and ``y``,
        where ``i < j``.

        EXAMPLES::

            sage: L = lie_algebras.SymplecticDerivation(QQ, 5)
            sage: L.bracket_on_basis([5,2,1], [5,1,1])
            0
            sage: L.bracket_on_basis([6,1], [3,1,1])
            -2*a1*a1*a3
            sage: L.bracket_on_basis([9,2,1], [4,1,1])
            -a1*a1*a1*a2
            sage: L.bracket_on_basis([5,5,2], [6,1,1])
            0
            sage: L.bracket_on_basis([5,5,5], [10,3])
            3*a3*a5*a5
            sage: L.bracket_on_basis([10,10,10], [5,3])
            -3*a3*b5*b5
        """
        g = self._g
        ret = {}
        one = self.base_ring().one()
        for i,xi in enumerate(x):
            for j,yj in enumerate(y):
                # The symplectic form will be 0
                if (xi <= g and yj <= g) or (xi > g and yj > g):
                    continue
                if xi <= g and yj > g:
                    if xi != yj - g:
                        continue
                    m = _Partitions(sorted(x[:i] + x[i+1:] + y[:j] + y[j+1:], reverse=True))
                    if m in ret:
                        ret[m] += one
                    else:
                        ret[m] = one
                else: # if ci > g and yj <= g:
                    if xi - g != yj:
                        continue
                    m = _Partitions(sorted(x[:i] + x[i+1:] + y[:j] + y[j+1:], reverse=True))
                    if m in ret:
                        ret[m] -= one
                    else:
                        ret[m] = -one
        return self._from_dict(ret, remove_zeros=True)

    def _an_element_(self):
        r"""
        Return an element of ``self``.

        EXAMPLES::

            sage: L = lie_algebras.SymplecticDerivation(QQ, 5)
            sage: L.an_element()
            a1*a2 - 1/2*a1*a2*a2*a5 + a1*a1*a2*b1*b4
        """
        d = self.monomial
        return (
                 d(_Partitions([2,1]))
                 - self.base_ring().an_element() * d(_Partitions([5,2,2,1]))
                 + d(_Partitions([2*self._g-1, self._g+1, 2, 1, 1]))
                )

    def some_elements(self):
        r"""
        Return some elements of ``self``.

        EXAMPLES::

            sage: L = lie_algebras.SymplecticDerivation(QQ, 5)
            sage: L.some_elements()
            [a1*a2, b1*b3, a1*a1*a2, b3*b4, a1*a4*b3,
             a1*a2 - 1/2*a1*a2*a2*a5 + a1*a1*a2*b1*b4]
        """
        d = self.monomial
        g = self._g
        return [d(_Partitions([2,1])), d(_Partitions([g+3,g+1])), d(_Partitions([2,1,1])),
                d(_Partitions([2*g-1,2*g-2])), d(_Partitions([2*g-2,g-1,1])),
                self.an_element()]

    class Element(LieAlgebraElement):
        pass



# sage_setup: distribution = sagemath-modules
"""
Helper code for ternary quadratic forms
"""

# ****************************************************************************
#       Copyright (C) 2012 Gustavo Rama
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.arith.misc import gcd, inverse_mod, xgcd
from sage.matrix.constructor import matrix
from sage.misc.prandom import randint
from sage.rings.finite_rings.integer_mod import mod
from sage.rings.integer_ring import ZZ


def red_mfact(a, b):
    """
    Auxiliary function for reduction that finds the reduction factor of integers `a`, `b`.

    INPUT:

    - ``a``, ``b`` -- integers

    OUTPUT: integer

    EXAMPLES::

        sage: from sage.quadratic_forms.ternary import red_mfact
        sage: red_mfact(0, 3)
        0
        sage: red_mfact(-5, 100)
        9
    """

    if a:
        return (-b + abs(a))//(2*a)
    else:
        return 0


def _reduced_ternary_form_eisenstein_with_matrix(a1, a2, a3, a23, a13, a12):
    """
    Find the coefficients of the equivalent unique reduced ternary form according to the conditions
    of Dickson's "Studies in the Theory of Numbers", pp164-171, and the transformation matrix.
    See TernaryQF.is_eisenstein_reduced for the conditions.

    EXAMPLES::

        sage: from sage.quadratic_forms.ternary import _reduced_ternary_form_eisenstein_with_matrix
        sage: Q = TernaryQF([293, 315, 756, 908, 929, 522])
        sage: qr, M = _reduced_ternary_form_eisenstein_with_matrix(293, 315, 756, 908, 929, 522)
        sage: qr
        (1, 2, 2, -1, 0, -1)
        sage: M
        [ -54  137  -38]
        [ -23   58  -16]
        [  47 -119   33]
        sage: Qr = TernaryQF(qr)
        sage: Qr.is_eisenstein_reduced()
        True
        sage: Q(M) == Qr
        True
    """
    # M = identity_matrix(3)
    # M = matrix(ZZ, 3, [m11, m12, m13, m21, m22, m23, m31, m32, m33])
    m11, m12, m13, m21, m22, m23, m31, m32, m33 = 1, 0, 0, 0, 1, 0, 0, 0, 1

    loop = True

    while loop:

        # adjust
        v = a1 + a2 + a23 + a13 + a12
        if (v < 0):
            # M *= matrix(ZZ, 3, [1, 0, 1, 0, 1, 1, 0, 0, 1])
            m13 = m11 + m12 + m13
            m23 = m21 + m22 + m23
            m33 = m31 + m32 + m33
            a3 += v
            a23 += a12 + 2*a2
            a13 += a12 + 2*a1

        # cuadred 12
        m = red_mfact(a1, a12)
        # M *= matrix(ZZ, 3, [1, m, 0, 0, 1, 0, 0, 0, 1])
        m12 = m*m11 + m12
        m22 = m*m21 + m22
        m32 = m*m31 + m32
        t = a1*m
        a12 += t
        a2 += a12*m
        a12 += t
        a23 += a13*m

        # cuadred 23
        m = red_mfact(a2, a23)
        # M *= matrix(ZZ, 3, [1, 0, 0, 0, 1, m, 0, 0, 1])
        m13 = m*m12 + m13
        m23 = m*m22 + m23
        m33 = m*m32 + m33
        t = a2*m
        a23 += t
        a3 += a23*m
        a23 += t
        a13 += a12*m

        # cuadred 13
        m = red_mfact(a1, a13)
        # M *= matrix(ZZ, 3, [1, 0, m, 0, 1, 0, 0, 0, 1])
        m13 = m*m11 + m13
        m23 = m*m21 + m23
        m33 = m*m31 + m33
        t = a1*m
        a13 += t
        a3 += a13*m
        a13 += t
        a23 += a12*m

        # order 12
        if a1 > a2 or (a1 == a2 and abs(a23) > abs(a13)):
            # M *= matrix(ZZ, 3, [0, -1, 0, -1, 0, 0, 0, 0, -1])
            m11, m12, m13 = [-m12, -m11, -m13]
            m21, m22, m23 = [-m22, -m21, -m23]
            m31, m32, m33 = [-m32, -m31, -m33]
            a1, a2 = [a2, a1]
            a13, a23 = [a23, a13]

        # order 23
        if a2 > a3 or (a2 == a3 and abs(a13) > abs(a12)):
            # M *= matrix(ZZ, 3, [-1, 0, 0, 0, 0, -1, 0, -1, 0])
            m11, m12, m13 = [-m11, -m13, -m12]
            m21, m22, m23 = [-m21, -m23, -m22]
            m31, m32, m33 = [-m31, -m33, -m32]
            a2, a3 = [a3, a2]
            a13, a12 = [a12, a13]

        # order 12
        if a1 > a2 or (a1 == a2 and abs(a23) > abs(a13)):
            # M *= matrix(ZZ, 3, [0, -1, 0, -1, 0, 0, 0, 0, -1])
            m11, m12, m13 = [-m12, -m11, -m13]
            m21, m22, m23 = [-m22, -m21, -m23]
            m31, m32, m33 = [-m32, -m31, -m33]
            a1, a2 = [a2, a1]
            a13, a23 = [a23, a13]

        # signs
        if a23*a13*a12 > 0:
            # a23, a13, a12 positive

            if (a23 < 0):
                # M *= diagonal_matrix([-1, 1, 1])
                m11 = -m11
                m21 = -m21
                m31 = -m31
                a23 = -a23
            if (a13 < 0):
                # M *= diagonal_matrix([1, -1, 1])
                m12 = -m12
                m22 = -m22
                m32 = -m32
                a13 = -a13
            if (a12 < 0):
                # M *= diagonal_matrix([1, 1, -1])
                m13 = -m13
                m23 = -m23
                m33 = -m33
                a12 = -a12

        else:
            # a23, a13, a12 nonpositive

            s1, s2, s3 = [a23 > 0, a13 > 0, a12 > 0]
            if (s1 + s2 + s3) % 2:
                if a23 == 0:
                    s1 = 1
                else:
                    if a13 == 0:
                        s2 = 1
                    else:
                        if (a12 == 0):
                            s3 = 1
            if s1:
                # M *= diagonal_matrix([-1, 1, 1])
                m11 = -m11
                m21 = -m21
                m31 = -m31
                a23 = -a23
            if s2:
                # M *= diagonal_matrix([1, -1, 1])
                m12 = -m12
                m22 = -m22
                m32 = -m32
                a13 = -a13
            if s3:
                # M *= diagonal_matrix([1, 1, -1])
                m13 = -m13
                m23 = -m23
                m33 = -m33
                a12 = -a12

        loop = not (abs(a23) <= a2 and abs(a13) <= a1 and abs(a12) <= a1 and a1 + a2 + a23 + a13 + a12 >= 0)

    # adj 3
    if a1 + a2 + a23 + a13 + a12 == 0 and 2*a1 + 2*a13 + a12 > 0:
        # M *= matrix(ZZ, 3, [-1, 0, 1, 0, -1, 1, 0, 0, 1])
        m11, m12, m13 = [-m11, -m12, m11 + m12 + m13]
        m21, m22, m23 = [-m21, -m22, m21 + m22 + m23]
        m31, m32, m33 = [-m31, -m32, m31 + m32 + m33]
        # a3 += a1+a2+a23+a13+a12 = 0
        a23 = -2*a2 - a23 - a12
        a13 = -2*a1 - a13 - a12

    # adj 5.12
    if a1 == -a12 and a13 != 0:
        # M *= matrix(ZZ, 3, [-1, -1, 0, 0, -1, 0, 0, 0, 1])
        m11, m12 = [-m11, -m11 - m12]
        m21, m22 = [-m21, -m21 - m22]
        m31, m32 = [-m31, -m31 - m32]
        # a2 += a1 + a12 = 0
        a23 = -a23 - a13
        a13 = -a13
        a12 = -a12  # = 2*a1 + a12

    # adj 5.13
    if a1 == -a13 and a12 != 0:
        # M *= matrix(ZZ, 3, [-1, 0, -1, 0, 1, 0, 0, 0, -1])
        m11, m13 = [-m11, -m11 - m13]
        m21, m23 = [-m21, -m21 - m23]
        m31, m33 = [-m31, -m31 - m33]
        # a3 += a1 + a13 = 0
        a23 = -a23 - a12
        a13 = -a13  # = 2*a1 + a13
        a12 = -a12

    # adj 5.23
    if a2 == -a23 and a12 != 0:
        # M *= matrix(ZZ, 3, [1, 0, 0, 0, -1, -1, 0, 0, -1])
        m12, m13 = [-m12, -m12 - m13]
        m22, m23 = [-m22, -m22 - m23]
        m32, m33 = [-m32, -m32 - m33]
        # a3 += a2 + a23 = 0
        a23 = -a23  # = 2*a2 + a23
        a13 = -a13 - a12
        a12 = -a12

    # adj 4.12
    if a1 == a12 and a13 > 2*a23:
        # M *= matrix(ZZ, 3, [-1, -1, 0, 0, 1, 0, 0, 0, -1])
        m11, m12, m13 = [-m11, -m11 + m12, -m13]
        m21, m22, m23 = [-m21, -m21 + m22, -m23]
        m31, m32, m33 = [-m31, -m31 + m32, -m33]
        # a2 += a1 - a12 = 0
        a23 = -a23 + a13
        # a12 = 2*a1 - a12

    # adj 4.13
    if a1 == a13 and a12 > 2*a23:
        # M *= matrix(ZZ, 3, [-1, 0, -1, 0, -1, 0, 0, 0, 1])
        m11, m12, m13 = [-m11, -m12, -m11 + m13]
        m21, m22, m23 = [-m21, -m22, -m21 + m23]
        m31, m32, m33 = [-m31, -m32, -m31 + m33]
        # a3 += a1 - a13 = 0
        a23 = -a23 + a12
        # a13 = 2*a1 - a13

    # adj 4.23
    if a2 == a23 and a12 > 2*a13:
        # M *= matrix(ZZ, 3, [-1, 0, 0, 0, -1, -1, 0, 0, 1])
        m11, m12, m13 = [-m11, -m12, -m12 + m13]
        m21, m22, m23 = [-m21, -m22, -m22 + m23]
        m31, m32, m33 = [-m31, -m32, -m32 + m33]
        # a3 += a2 - a23 = 0
        # a23 = 2*a2 - a23
        a13 = -a13 + a12

    # order 12
    if a1 == a2 and abs(a23) > abs(a13):
        # M *= matrix(ZZ, 3, [0, -1, 0, -1, 0, 0, 0, 0, -1])
        m11, m12, m13 = [-m12, -m11, -m13]
        m21, m22, m23 = [-m22, -m21, -m23]
        m31, m32, m33 = [-m32, -m31, -m33]
        a1, a2 = [a2, a1]
        a13, a23 = [a23, a13]

    # order 23
    if a2 == a3 and abs(a13) > abs(a12):
        # M *= matrix(ZZ, 3, [-1, 0, 0, 0, 0, -1, 0, -1, 0])
        m11, m12, m13 = [-m11, -m13, -m12]
        m21, m22, m23 = [-m21, -m23, -m22]
        m31, m32, m33 = [-m31, -m33, -m32]
        a13, a12 = [a12, a13]

    # order 12
    if a1 == a2 and abs(a23) > abs(a13):
        # M *= matrix(ZZ, 3, [0, -1, 0, -1, 0, 0, 0, 0, -1])
        m11, m12, m13 = [-m12, -m11, -m13]
        m21, m22, m23 = [-m22, -m21, -m23]
        m31, m32, m33 = [-m32, -m31, -m33]
        a13, a23 = [a23, a13]

    return (a1, a2, a3, a23, a13, a12), \
        matrix(ZZ, 3, (m11, m12, m13, m21, m22, m23, m31, m32, m33))


def _reduced_ternary_form_eisenstein_without_matrix(a1, a2, a3, a23, a13, a12):
    """
    Find the coefficients of the equivalent unique reduced ternary form according to the conditions
    of Dickson's "Studies in the Theory of Numbers", pp164-171.
    See TernaryQF.is_eisenstein_reduced for the conditions.

    EXAMPLES::

        sage: from sage.quadratic_forms.ternary import _reduced_ternary_form_eisenstein_without_matrix
        sage: Q = TernaryQF([293, 315, 756, 908, 929, 522])
        sage: qr = _reduced_ternary_form_eisenstein_without_matrix(293, 315, 756, 908, 929, 522)
        sage: qr
        (1, 2, 2, -1, 0, -1)
        sage: Qr = TernaryQF(qr)
        sage: Qr.is_eisenstein_reduced()
        True
    """
    loop = True

    while loop:

        # adjust
        v = a1+a2+a23+a13+a12
        if v < 0:
            a3 += v
            a23 += a12+2*a2
            a13 += a12+2*a1

        # cuadred 12
        m = red_mfact(a1, a12)
        t = a1*m
        a12 += t
        a2 += a12*m
        a12 += t
        a23 += a13*m

        # cuadred 23
        m = red_mfact(a2, a23)
        t = a2*m
        a23 += t
        a3 += a23*m
        a23 += t
        a13 += a12*m

        # cuadred 13
        m = red_mfact(a1, a13)
        t = a1*m
        a13 += t
        a3 += a13*m
        a13 += t
        a23 += a12*m

        # order 12
        if a1 > a2 or (a1 == a2 and abs(a23) > abs(a13)):
            [a1, a2] = [a2, a1]
            [a13, a23] = [a23, a13]

        # order 23
        if a2 > a3 or (a2 == a3 and abs(a13) > abs(a12)):
            [a2, a3] = [a3, a2]
            [a13, a12] = [a12, a13]

        # order 12
        if a1 > a2 or (a1 == a2 and abs(a23) > abs(a13)):
            [a1, a2] = [a2, a1]
            [a13, a23] = [a23, a13]

        # signs
        if a23*a13*a12 > 0:
            # a23, a13, a12 positive

            if a23 < 0:
                a23 = -a23
            if a13 < 0:
                a13 = -a13
            if a12 < 0:
                a12 = -a12

        else:
            # a23, a13, a12 nonpositive

            [s1, s2, s3] = [a23 > 0, a13 > 0, a12 > 0]
            if (s1+s2+s3) % 2:
                if a23 == 0:
                    s1 = 1
                elif a13 == 0:
                    s2 = 1
                elif a12 == 0:
                    s3 = 1
            if s1:
                a23 = -a23
            if s2:
                a13 = -a13
            if s3:
                a12 = -a12

        loop = not (abs(a23) <= a2 and abs(a13) <= a1 and abs(a12) <= a1 and a1+a2+a23+a13+a12 >= 0)

    # adj 3
    if a1+a2+a23+a13+a12 == 0 and 2*a1+2*a13+a12 > 0:
        # a3 += a1+a2+a23+a13+a12
        a23 = -2*a2-a23-a12
        a13 = -2*a1-a13-a12

    # adj 5.12
    if a1 == -a12 and a13 != 0:
        # a2 += a1+a12
        a23 = -a23-a13
        a13 = -a13
        a12 = -a12  # = 2*a1+a12

    # adj 5.13
    if a1 == -a13 and a12 != 0:
        # a3 += a1+a13
        a23 = -a23-a12
        a13 = -a13  # = 2*a1+a13
        a12 = -a12

    # adj 5.23
    if a2 == -a23 and a12 != 0:
        # a3 += a2+a23
        a23 = -a23  # = 2*a2+a23
        a13 = -a13-a12
        a12 = -a12

    # adj 4.12
    if a1 == a12 and a13 > 2*a23:
        # a 2 += a1-a12
        a23 = -a23 + a13
        # a12 = 2*a1 - a12

    # adj 4.13
    if a1 == a13 and a12 > 2*a23:
        # a3 += a1-a13
        a23 = -a23 + a12
        # a13 = 2*a1 - a13

    # adj 4.23
    if a2 == a23 and a12 > 2*a13:
        # a3 += a2-a23
        # a23 = 2*a2 - a23
        a13 = -a13 + a12

    # order 12
    if a1 == a2 and abs(a23) > abs(a13):
        [a1, a2] = [a2, a1]
        [a13, a23] = [a23, a13]

    # order 23
    if a2 == a3 and abs(a13) > abs(a12):
        [a13, a12] = [a12, a13]

    # order 12
    if a1 == a2 and abs(a23) > abs(a13):
        [a13, a23] = [a23, a13]

    return a1, a2, a3, a23, a13, a12


def primitivize(long long v0, long long v1, long long v2, p):
    """
    Given a 3-tuple `v` not singular mod `p`, return a primitive 3-tuple version of `v` mod `p`.

    EXAMPLES::

        sage: from sage.quadratic_forms.ternary import primitivize
        sage: primitivize(12, 13, 14, 5)
        (3, 2, 1)
        sage: primitivize(12, 13, 15, 5)
        (4, 1, 0)
    """
    if v2 % p:
        v2_inv = inverse_mod(v2, p)
        return v2_inv*v0 % p, v2_inv*v1 % p, 1
    elif v1 % p:
        return inverse_mod(v1, p)*v0 % p, 1, 0
    else:
        return 1, 0, 0


def evaluate(a, b, c, r, s, t, v):
    """
    Function to evaluate the ternary quadratic form `(a, b, c, r, s, t)` in a 3-tuple `v`.

    EXAMPLES::

        sage: from sage.quadratic_forms.ternary import evaluate
        sage: Q = TernaryQF([1, 2, 3, -1, 0, 0])
        sage: v = (1, -1, 19)
        sage: Q(v)
        1105
        sage: evaluate(1, 2, 3, -1, 0, 0, v)
        1105
    """
    return a*v[0]**2+b*v[1]**2+c*v[2]**2+r*v[2]*v[1]+s*v[2]*v[0]+t*v[1]*v[0]


def _find_zeros_mod_p_2(a, b, c, r, s, t):
    """
    Function to find the zeros mod 2 of a ternary quadratic form.

    EXAMPLES::

        sage: Q = TernaryQF([1, 2, 2, -1, 0, 0])
        sage: from sage.quadratic_forms.ternary import _find_zeros_mod_p_2
        sage: zeros = _find_zeros_mod_p_2(1, 2, 2, -1, 0, 0)
        sage: zeros
        [(0, 1, 0), (0, 0, 1), (1, 1, 1)]
        sage: Q((0, 1, 0))
        2
        sage: Q((0, 0, 1))
        2
        sage: Q((1, 1, 1))
        4
    """
    zeros = []
    v = (1, 0, 0)
    if evaluate(a, b, c, r, s, t, v) % 2 == 0:
        zeros.append(v)
    for i in range(2):
        v = (i, 1, 0)
        if evaluate(a, b, c, r, s, t, v) % 2 == 0:
            zeros.append(v)
    for i in range(2):
        for j in range(2):
            v = (i, j, 1)
            if evaluate(a, b, c, r, s, t, v) % 2 == 0:
                zeros.append(v)
    return zeros


def pseudorandom_primitive_zero_mod_p(a, b, c, r, s, t, p):
    """
    Find a zero of the form `(a, b, 1)` of the ternary quadratic form given by the coefficients `(a, b, c, r, s, t)`
    mod `p`, where `p` is a odd prime that doesn't divide the discriminant.

    EXAMPLES::

        sage: Q = TernaryQF([1, 2, 2, -1, 0, 0])
        sage: p = 1009
        sage: from sage.quadratic_forms.ternary import pseudorandom_primitive_zero_mod_p
        sage: v = pseudorandom_primitive_zero_mod_p(1, 2, 2, -1, 0, 0, p)               # needs sage.libs.pari
        sage: v[2]                                                                      # needs sage.libs.pari
        1
        sage: Q(v)%p                                                                    # needs sage.libs.pari
        0
    """
    # [a, b, c, r, s, t] = Q.coefficients()
    while True:

        r1 = randint(0, p-1)
        r2 = randint(0, p-1)
        alpha = (b*r1**2+t*r1+a) % p
        if alpha != 0:

            beta = (2*b*r1*r2+t*r2+r*r1+s) % p
            gamma = (b*r2**2+r*r2+c) % p
            disc = beta**2-4*alpha*gamma
            if mod(disc, p).is_square():

                z = (-beta+mod(disc, p).sqrt().lift())*(2*alpha).inverse_mod(p)
                return z % p, (r1*z+r2) % p, 1


def _find_zeros_mod_p_odd(long long a, long long b, long long c, long long r, long long s, long long t, long long p, v):
    """
    Find the zeros mod `p`, where `p` is an odd prime, of a ternary quadratic form given by its coefficients and a given zero of the form `v`.

    The prime `p` does not divide the discriminant of the form.

    EXAMPLES::

        sage: from sage.quadratic_forms.ternary import _find_zeros_mod_p_odd
        sage: Q = TernaryQF([1, 2, 2, -1, 0, 0])
        sage: p = 1009
        sage: v = (817, 974, 1)
        sage: Q(v)%1009
        0
        sage: zeros_1009 = _find_zeros_mod_p_odd(1, 2, 2, -1, 0, 0, 1009, v)
        sage: len(zeros_1009)
        1010
        sage: zeros_1009.sort()
        sage: zeros_1009[0]
        (0, 32, 1)
        sage: Q((0, 32, 1))
        2018
    """

    cdef long long a_i
    cdef long long c_i
    cdef long long a_inf
    cdef long long c_inf
    cdef long long i
    cdef long long l
    cdef long long x0
    cdef long long y0

    zeros = [v]
    x0 = v[0]
    y0 = v[1]
    more = False
    for i in range(p):
        a_i = (a*x0**2+b*i**2-2*b*i*y0+b*y0**2-t*i*x0+t*x0*y0-2*a*x0+t*i-t*y0+s*x0-r*i+r*y0+a+c-s) % p
        # b_i=(-2*b*i**2+2*b*i*y0+t*i*x0+2*a*x0-2*t*i+t*y0+r*i-2*a+s)%p
        if a_i == 0:
            w = ((x0-1) % p, (y0-i) % p, 1)
            if w == v:
                more = True
            else:
                zeros.append(w)
        else:
            c_i = (b*i**2+t*i+a) % p
            l = c_i*ZZ(a_i).inverse_mod(p)
            w = primitivize(l*(x0-1)+1, l*(y0-i)+i, l, p)
            if w[0] == v[0] and w[1] == v[1] and w[2] == v[2]:
                more = True
            else:
                zeros.append(w)
    if more:
        a_inf = (a*x0**2+b*y0**2+t*x0*y0-2*b*y0-t*x0+s*x0+r*y0+b+c-r) % p
        # b_inf=(2*b*y0+t*x0-2*b+r)%p
        if a_inf == 0:
            w = (x0 % p, (y0-1) % p, 1)
            zeros.append(w)
        else:
            c_inf = b % p
            l = c_inf*ZZ(a_inf).inverse_mod(p)
            w = primitivize(l*x0, l*(y0-1)+1, l, p)
            zeros.append(w)

    return zeros


def _find_zeros_mod_p(a, b, c, r, s, t, p):
    """
    Find the zeros mod `p` of the ternary quadratic form.

    The quadratic form is given by the coefficients `(a, b, c, r, s, t)`,
    and `p` is a prime that does not divide the discriminant of the form.

    EXAMPLES::

        sage: # needs sage.libs.pari
        sage: from sage.quadratic_forms.ternary import _find_zeros_mod_p
        sage: Q = TernaryQF([1, 2, 2, -1, 0, 0])
        sage: p = 1009
        sage: zeros_1009 = _find_zeros_mod_p(1, 2, 2, -1, 0, 0, p)
        sage: len(zeros_1009)
        1010
        sage: zeros_1009.sort()
        sage: zeros_1009[0]
        (0, 32, 1)
        sage: Q((0, 32, 1))
        2018
        sage: zeros_2 = _find_zeros_mod_p(1, 2, 2, -1, 0, 0, 2)
        sage: zeros_2
        [(0, 1, 0), (0, 0, 1), (1, 1, 1)]
    """
    if p == 2:
        return _find_zeros_mod_p_2(a, b, c, r, s, t)
    v = pseudorandom_primitive_zero_mod_p(a, b, c, r, s, t, p)
    return _find_zeros_mod_p_odd(a, b, c, r, s, t, p, v)


def _find_all_ternary_qf_by_level_disc(long long N, long long d):
    """
    Find the coefficients of all the reduced ternary quadratic forms given its discriminant d and level N.
    If N|4d and d|N^2, then it may be some forms with that discriminant and level.

    EXAMPLES::

        sage: from sage.quadratic_forms.ternary import _find_all_ternary_qf_by_level_disc
        sage: _find_all_ternary_qf_by_level_disc(44, 11)
        [(1, 1, 3, 0, -1, 0), (1, 1, 4, 1, 1, 1)]
        sage: _find_all_ternary_qf_by_level_disc(44, 11^2 * 16)
        [(3, 15, 15, -14, -2, -2), (4, 11, 12, 0, -4, 0)]
        sage: Q = TernaryQF([1, 1, 3, 0, -1, 0])
        sage: Q.is_eisenstein_reduced()
        True
        sage: Q.reciprocal_reduced()
        Ternary quadratic form with integer coefficients:
        [4 11 12]
        [0 -4 0]
        sage: _find_all_ternary_qf_by_level_disc(44, 22)
        []
        sage: _find_all_ternary_qf_by_level_disc(44, 33)
        Traceback (most recent call last):
        ...
        ValueError: There are no ternary forms of this level and discriminant
    """
    cdef long long a
    cdef long long b
    cdef long long c
    cdef long long r
    cdef long long s
    cdef long long t
    cdef long long m
    cdef long long g
    cdef long long u
    cdef long long v
    cdef long long g1
    cdef long long m_q
    cdef double a_max
    cdef double r_max
    cdef double b_max

    l = []

    if (4*d) % N:
        raise ValueError("There are no ternary forms of this level and discriminant")
    else:
        m = 4*d//N

    if (N**2) % d:
        raise ValueError("There are no ternary forms of this level and discriminant")

    a = 1
    a_max = (d/2.)**(1/3.)
    while a <= a_max:

        g, u, v = xgcd(4*a, m)
        g1 = (ZZ(g).squarefree_part()*g).sqrtrem()[0]
        t = 0
        while t <= a:

            alpha = max(a, m/4/a)
            b = alpha+((((u*t*t//g)-alpha)) % (m//g))
            b_max = (d/2.0/a)**(1/2.)
            while b <= b_max:

                s = 0
                beta = g//gcd(g, 2*t)
                while s <= a:

                    r = -b+((((2*s*t*u//g)+b)) % (m//g))
                    if s*t == 0:
                        r_max = 0
                    else:
                        r_max = b
                    while r <= r_max:

                        if (d-r*s*t+a*r*r+b*s*s) % (4*a*b-t**2) == 0:

                            c = (d-r*s*t+a*r*r+b*s*s)//(4*a*b-t**2)
                            if r <= 0:
                                is_reduced = True
                                if r < -b:
                                    is_reduced = False
                                elif not (b <= c and 0 <= a+b+r-s-t):
                                    is_reduced = False
                                elif a == b and abs(r) > abs(s):
                                    is_reduced = False
                                elif b == c and abs(s) > abs(t):
                                    is_reduced = False
                                elif a+b+r-s-t == 0 and 2*a-2*s-t > 0:
                                    is_reduced = False
                                elif a == t and s != 0:
                                    is_reduced = False
                                elif a == s and t != 0:
                                    is_reduced = False
                                elif b == -r and t != 0:
                                    is_reduced = False
                                if is_reduced:
                                    m_q = gcd((4*b*c-r**2, 4*a*c-s**2, 4*a*b-t**2, 2*s*t-4*a*r, -2*r*t+4*b*s, -2*r*s+4*c*t))
                                    if m_q == m:
                                        l.append((ZZ(a), ZZ(b), ZZ(c), ZZ(r), ZZ(-s), ZZ(-t)))
                            else:
                                is_reduced = True
                                if not (b <= c and 0 <= a+b+r+s+t):
                                    is_reduced = False
                                elif a == b and abs(r) > abs(s):
                                    is_reduced = False
                                elif b == c and abs(s) > abs(t):
                                    is_reduced = False
                                elif a+b+r+s+t == 0 and 2*a+2*s+t > 0:
                                    is_reduced = False
                                elif a == t and s > 2*r:
                                    is_reduced = False
                                elif a == s and t > 2*r:
                                    is_reduced = False
                                elif b == r and t > 2*s:
                                    is_reduced = False
                                if is_reduced:
                                    m_q = gcd((4*b*c-r**2, 4*a*c-s**2, 4*a*b-t**2, 2*s*t-4*a*r, 2*r*t-4*b*s, 2*r*s-4*c*t))
                                    if m_q == m:
                                        l.append((ZZ(a), ZZ(b), ZZ(c), ZZ(r), ZZ(s), ZZ(t)))
                        r += (m//g)
                    s += beta
                b += m//g
            t += g1

        a += 1
    return l


def _find_a_ternary_qf_by_level_disc(long long N, long long d):
    """
    Find the coefficients of a reduced ternary quadratic form given its discriminant d and level N.
    If N|4d and d|N^2, then it may be a form with that discriminant and level.

    EXAMPLES::

        sage: from sage.quadratic_forms.ternary import _find_a_ternary_qf_by_level_disc
        sage: _find_a_ternary_qf_by_level_disc(44, 11)
        (1, 1, 3, 0, -1, 0)
        sage: _find_a_ternary_qf_by_level_disc(44, 11^2 * 16)
        (3, 15, 15, -14, -2, -2)
        sage: Q = TernaryQF([1, 1, 3, 0, -1, 0])
        sage: Q.is_eisenstein_reduced()
        True
        sage: Q.level()
        44
        sage: Q.disc()
        11
        sage: _find_a_ternary_qf_by_level_disc(44, 22)
        sage: _find_a_ternary_qf_by_level_disc(44, 33)
        Traceback (most recent call last):
        ...
        ValueError: There are no ternary forms of this level and discriminant
    """
    cdef long long a
    cdef long long b
    cdef long long c
    cdef long long r
    cdef long long s
    cdef long long t
    cdef long long m
    cdef long long g
    cdef long long u
    cdef long long v
    cdef long long g1
    cdef long long m_q
    cdef double a_max
    cdef double r_max
    cdef double b_max

    if (4*d) % N:
        raise ValueError("There are no ternary forms of this level and discriminant")
    else:
        m = 4*d//N

    if (N**2) % d:
        raise ValueError("There are no ternary forms of this level and discriminant")

    a = 1
    a_max = (d/2.)**(1/3.)
    while a <= a_max:

        g, u, v = xgcd(4*a, m)
        g1 = (ZZ(g).squarefree_part()*g).sqrtrem()[0]
        t = 0
        while t <= a:

            alpha = max(a, m/4/a)
            b = alpha+((((u*t*t//g)-alpha)) % (m//g))
            b_max = (d/2.0/a)**(1/2.)
            while b <= b_max:

                s = 0
                beta = g//gcd(g, 2*t)
                while s <= a:

                    r = -b+((((2*s*t*u//g)+b)) % (m//g))
                    if s*t == 0:
                        r_max = 0
                    else:
                        r_max = b
                    while r <= r_max:

                        if (d-r*s*t+a*r*r+b*s*s) % (4*a*b-t**2) == 0:

                            c = (d-r*s*t+a*r*r+b*s*s)//(4*a*b-t**2)
                            if r <= 0:
                                is_reduced = True
                                if r < -b:
                                    is_reduced = False
                                elif not (b <= c and 0 <= a+b+r-s-t):
                                    is_reduced = False
                                elif a == b and abs(r) > abs(s):
                                    is_reduced = False
                                elif b == c and abs(s) > abs(t):
                                    is_reduced = False
                                elif a+b+r-s-t == 0 and 2*a-2*s-t > 0:
                                    is_reduced = False
                                elif a == t and s != 0:
                                    is_reduced = False
                                elif a == s and t != 0:
                                    is_reduced = False
                                elif b == -r and t != 0:
                                    is_reduced = False
                                if is_reduced:
                                    m_q = gcd((4*b*c-r**2, 4*a*c-s**2, 4*a*b-t**2, 2*s*t-4*a*r, -2*r*t+4*b*s, -2*r*s+4*c*t))
                                    if m_q == m:
                                        return ZZ(a), ZZ(b), ZZ(c), ZZ(r), ZZ(-s), ZZ(-t)
                            else:
                                is_reduced = True
                                if not (b <= c and 0 <= a+b+r+s+t):
                                    is_reduced = False
                                elif a == b and abs(r) > abs(s):
                                    is_reduced = False
                                elif b == c and abs(s) > abs(t):
                                    is_reduced = False
                                elif a+b+r+s+t == 0 and 2*a+2*s+t > 0:
                                    is_reduced = False
                                elif a == t and s > 2*r:
                                    is_reduced = False
                                elif a == s and t > 2*r:
                                    is_reduced = False
                                elif b == r and t > 2*s:
                                    is_reduced = False
                                if is_reduced:
                                    m_q = gcd((4*b*c-r**2, 4*a*c-s**2, 4*a*b-t**2, 2*s*t-4*a*r, 2*r*t-4*b*s, 2*r*s-4*c*t))
                                    if m_q == m:
                                        return ZZ(a), ZZ(b), ZZ(c), ZZ(r), ZZ(s), ZZ(t)
                        r += (m//g)
                    s += beta
                b += m//g
            t += g1

        a += 1


def extend(v):
    """
    Return the coefficients of a matrix `M` such that `M` has determinant gcd(`v`) and the first column is `v`.

    EXAMPLES::

        sage: from sage.quadratic_forms.ternary import extend
        sage: v = (6, 4, 12)
        sage: m = extend(v)
        sage: M = matrix(3, m); M
        [ 6  1  0]
        [ 4  1  0]
        [12  0  1]
        sage: M.det()
        2
        sage: v = (-12, 20, 30)
        sage: m = extend(v)
        sage: M = matrix(3, m)
        sage: M
        [-12   1   0]
        [ 20  -2   1]
        [ 30   0  -7]
        sage: M.det()
        2

    TESTS::

        sage: M = matrix(3, extend( (0, 0, 1) ))
        sage: M.det()
        1
        sage: M.column(0)
        (0, 0, 1)
        sage: M = matrix(3, extend( (0, 0, -2) ))
        sage: M.det()
        2
        sage: M.column(0)
        (0, 0, -2)
    """
    if v[0] == v[1] == 0:
        if v[2] < 0:
            return v[0], 0, 1, v[1], 1, 0, v[2], 0, 0
        else:
            return v[0], 1, 0, v[1], 0, 1, v[2], 0, 0

    b1 = xgcd(v[0], v[1])
    b2 = xgcd(b1[1], b1[2])
    b3 = xgcd(b1[0], v[2])

    return v[0], -b1[2], -b2[1]*b3[2], v[1], b1[1], -b2[2]*b3[2], v[2], 0, b3[1]


def _find_p_neighbor_from_vec(a, b, c, r, s, t, p, v, mat=False):
    """
    Find the coefficients of the reduced equivalent of the p-neighbor
    of the ternary quadratic given by Q = (a, b, c, r, s, t)  form associated
    to a given vector v satisfying:

    1. Q(v) = 0  mod p

    2. v is a non-singular point of the conic Q(v) = 0 mod p.

    Reference:  Gonzalo Tornaria's Thesis, Thrm 3.5, p34.

    EXAMPLES::

        sage: from sage.quadratic_forms.ternary import _find_p_neighbor_from_vec
        sage: Q = TernaryQF([1, 3, 3, -2, 0, -1])
        sage: Q
        Ternary quadratic form with integer coefficients:
        [1 3 3]
        [-2 0 -1]
        sage: Q.disc()
        29
        sage: v = (9, 7, 1)
        sage: v in Q.find_zeros_mod_p(11)                                               # needs sage.libs.pari
        True
        sage: q11, M = _find_p_neighbor_from_vec(1, 3, 3, -2, 0, -1, 11, v, mat=True)
        sage: Q11 = TernaryQF(q11)
        sage: Q11
        Ternary quadratic form with integer coefficients:
        [1 2 4]
        [-1 -1 0]
        sage: M = matrix(3, M)
        sage: M
        [    -1  -5/11   7/11]
        [     0 -10/11   3/11]
        [     0  -3/11  13/11]
        sage: Q(M) == Q11
        True
    """
    v0, w0, u0, v1, w1, u1, v2, w2, u2 = extend(v)

    m00 = 2*a*v0**2 + 2*b*v1**2 + 2*c*v2**2 + 2*r*v1*v2 + 2*s*v0*v2 + 2*t*v0*v1
    m11 = 2*a*w0**2 + 2*b*w1**2 + 2*t*w0*w1
    m22 = 2*a*u0**2 + 2*b*u1**2 + 2*c*u2**2 + 2*r*u1*u2 + 2*s*u0*u2 + 2*t*u0*u1
    m01 = 2*a*v0*w0 + 2*b*v1*w1 + r*v2*w1 + s*v2*w0 + t*v0*w1 + t*v1*w0
    m02 = 2*a*u0*v0 + 2*b*u1*v1 + 2*c*u2*v2 + r*u1*v2 + r*u2*v1 + s*u0*v2 + s*u2*v0 + t*u0*v1 + t*u1*v0
    m12 = 2*a*u0*w0 + 2*b*u1*w1 + r*u2*w1 + s*u2*w0 + t*u0*w1 + t*u1*w0

    if m02 % p:

        m0 = (-m00/m02/2) % p**2
        m1 = (-m01/m02) % p

        b00 = m0**2*m22/p**2 + 2*m0*m02/p**2 + m00/p**2
        b11 = m1**2*m22 + 2*m1*m12 + m11
        b22 = m22*p**2
        b01 = m0*m1*m22/p + m0*m12/p + m02*m1/p + m01/p
        b02 = m0*m22 + m02
        b12 = m1*m22*p + m12*p

        if mat:
            q, Mr = _reduced_ternary_form_eisenstein_with_matrix(ZZ(b00/2), ZZ(b11/2), ZZ(b22/2), ZZ(b12), ZZ(b02), ZZ(b01))
            r00, r01, r02, r10, r11, r12, r20, r21, r22 = Mr.list()
            t00 = p*r20*u0 + (m0*u0/p + v0/p)*r00 + (m1*u0 + w0)*r10
            t01 = p*r21*u0 + (m0*u0/p + v0/p)*r01 + (m1*u0 + w0)*r11
            t02 = p*r22*u0 + (m0*u0/p + v0/p)*r02 + (m1*u0 + w0)*r12
            t10 = p*r20*u1 + (m0*u1/p + v1/p)*r00 + (m1*u1 + w1)*r10
            t11 = p*r21*u1 + (m0*u1/p + v1/p)*r01 + (m1*u1 + w1)*r11
            t12 = p*r22*u1 + (m0*u1/p + v1/p)*r02 + (m1*u1 + w1)*r12
            t20 = p*r20*u2 + (m0*u2/p + v2/p)*r00 + (m1*u2 + w2)*r10
            t21 = p*r21*u2 + (m0*u2/p + v2/p)*r01 + (m1*u2 + w2)*r11
            t22 = p*r22*u2 + (m0*u2/p + v2/p)*r02 + (m1*u2 + w2)*r12
            return q, (t00, t01, t02, t10, t11, t12, t20, t21, t22)
        else:
            return _reduced_ternary_form_eisenstein_without_matrix(ZZ(b00/2), ZZ(b11/2), ZZ(b22/2), ZZ(b12), ZZ(b02), ZZ(b01))

    if m01 % p:

        m0 = (-m00/m01/2) % p**2
        m1 = (-m02/m01) % p

        b00 = m0**2*m11/p**2 + 2*m0*m01/p**2 + m00/p**2
        b11 = m1**2*m11 + 2*m1*m12 + m22
        b22 = m11*p**2
        b01 = m0*m1*m11/p + m0*m12/p + m01*m1/p + m02/p
        b02 = m0*m11 + m01
        b12 = m1*m11*p + m12*p

        if mat:
            q, Mr = _reduced_ternary_form_eisenstein_with_matrix(ZZ(b00/2), ZZ(b11/2), ZZ(b22/2), ZZ(b12), ZZ(b02), ZZ(b01))
            r00, r01, r02, r10, r11, r12, r20, r21, r22 = Mr.list()
            t00 = p*r20*w0 + (m0*w0/p + v0/p)*r00 + (m1*w0 + u0)*r10
            t01 = p*r21*w0 + (m0*w0/p + v0/p)*r01 + (m1*w0 + u0)*r11
            t02 = p*r22*w0 + (m0*w0/p + v0/p)*r02 + (m1*w0 + u0)*r12
            t10 = p*r20*w1 + (m0*w1/p + v1/p)*r00 + (m1*w1 + u1)*r10
            t11 = p*r21*w1 + (m0*w1/p + v1/p)*r01 + (m1*w1 + u1)*r11
            t12 = p*r22*w1 + (m0*w1/p + v1/p)*r02 + (m1*w1 + u1)*r12
            t20 = p*r20*w2 + (m0*w2/p + v2/p)*r00 + (m1*w2 + u2)*r10
            t21 = p*r21*w2 + (m0*w2/p + v2/p)*r01 + (m1*w2 + u2)*r11
            t22 = p*r22*w2 + (m0*w2/p + v2/p)*r02 + (m1*w2 + u2)*r12
            return q, (t00, t01, t02, t10, t11, t12, t20, t21, t22)
        else:
            return _reduced_ternary_form_eisenstein_without_matrix(ZZ(b00/2), ZZ(b11/2), ZZ(b22/2), ZZ(b12), ZZ(b02), ZZ(b01))


def _basic_lemma_vec(a, b, c, r, s, t, n):
    """
    Find a vector v such that the ternary quadratic form given by (a, b, c, r, s, t) evaluated at v is
    coprime with n a prime or 1.

    EXAMPLES::

        sage: from sage.quadratic_forms.ternary import _basic_lemma_vec
        sage: Q = TernaryQF([5, 2, 3, -1, 0, 0])
        sage: v = _basic_lemma_vec(5, 2, 3, -1, 0, 0, 5)
        sage: v
        (0, 1, 0)
        sage: Q(v)
        2
    """
    if n == 1:
        return 0, 0, 0

    if a % n:
        return 1, 0, 0
    elif b % n:
        return 0, 1, 0
    elif c % n:
        return 0, 0, 1

    if r % n:
        return 0, 1, 1
    elif s % n:
        return 1, 0, 1
    elif t % n:
        return 1, 1, 0

    raise ValueError("not primitive form")


def _basic_lemma(a, b, c, r, s, t, n):
    """
    Find a number represented by the ternary quadratic form given by the coefficients (a, b, c, r, s, t)
    and coprime to the prime n.

    EXAMPLES::

        sage: from sage.quadratic_forms.ternary import _basic_lemma
        sage: Q = TernaryQF([5, 2, 3, -1, 0, 0])
        sage: _basic_lemma(5, 2, 3, -1, 0, 0, 5)
        2
    """
    if n == 1:
        return 0

    if a % n:
        return a
    elif b % n:
        return b
    elif c % n:
        return c

    if r % n:
        return b + c + r
    elif s % n:
        return a + c + s
    elif t % n:
        return a + b + t

    raise ValueError("not primitive form")



# sage_setup: distribution = sagemath-flint
# distutils: libraries = flint
# distutils: depends = flint/acb_elliptic.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/src/sage_setup/autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    void acb_elliptic_k(acb_t res, const acb_t m, slong prec) noexcept
    void acb_elliptic_k_jet(acb_ptr res, const acb_t m, slong len, slong prec) noexcept
    void _acb_elliptic_k_series(acb_ptr res, acb_srcptr m, slong mlen, slong len, slong prec) noexcept
    void acb_elliptic_k_series(acb_poly_t res, const acb_poly_t m, slong len, slong prec) noexcept
    void acb_elliptic_e(acb_t res, const acb_t m, slong prec) noexcept
    void acb_elliptic_pi(acb_t res, const acb_t n, const acb_t m, slong prec) noexcept
    void acb_elliptic_f(acb_t res, const acb_t phi, const acb_t m, int pi, slong prec) noexcept
    void acb_elliptic_e_inc(acb_t res, const acb_t phi, const acb_t m, int pi, slong prec) noexcept
    void acb_elliptic_pi_inc(acb_t res, const acb_t n, const acb_t phi, const acb_t m, int pi, slong prec) noexcept
    void acb_elliptic_rf(acb_t res, const acb_t x, const acb_t y, const acb_t z, int flags, slong prec) noexcept
    void acb_elliptic_rg(acb_t res, const acb_t x, const acb_t y, const acb_t z, int flags, slong prec) noexcept
    void acb_elliptic_rj(acb_t res, const acb_t x, const acb_t y, const acb_t z, const acb_t p, int flags, slong prec) noexcept
    void acb_elliptic_rj_carlson(acb_t res, const acb_t x, const acb_t y, const acb_t z, const acb_t p, int flags, slong prec) noexcept
    void acb_elliptic_rj_integration(acb_t res, const acb_t x, const acb_t y, const acb_t z, const acb_t p, int flags, slong prec) noexcept
    void acb_elliptic_rc1(acb_t res, const acb_t x, slong prec) noexcept
    void acb_elliptic_p(acb_t res, const acb_t z, const acb_t tau, slong prec) noexcept
    void acb_elliptic_p_prime(acb_t res, const acb_t z, const acb_t tau, slong prec) noexcept
    void acb_elliptic_p_jet(acb_ptr res, const acb_t z, const acb_t tau, slong len, slong prec) noexcept
    void _acb_elliptic_p_series(acb_ptr res, acb_srcptr z, slong zlen, const acb_t tau, slong len, slong prec) noexcept
    void acb_elliptic_p_series(acb_poly_t res, const acb_poly_t z, const acb_t tau, slong len, slong prec) noexcept
    void acb_elliptic_invariants(acb_t g2, acb_t g3, const acb_t tau, slong prec) noexcept
    void acb_elliptic_roots(acb_t e1, acb_t e2, acb_t e3, const acb_t tau, slong prec) noexcept
    void acb_elliptic_inv_p(acb_t res, const acb_t z, const acb_t tau, slong prec) noexcept
    void acb_elliptic_zeta(acb_t res, const acb_t z, const acb_t tau, slong prec) noexcept
    void acb_elliptic_sigma(acb_t res, const acb_t z, const acb_t tau, slong prec) noexcept



# sage_setup: distribution = sagemath-modules
"""
Partition backtrack functions for matrices

EXAMPLES::

    sage: import sage.groups.perm_gps.partn_ref.refinement_matrices

REFERENCE:

- [1] McKay, Brendan D. Practical Graph Isomorphism. Congressus Numerantium,
  Vol. 30 (1981), pp. 45-87.

- [2] Leon, Jeffrey. Permutation Group Algorithms Based on Partitions, I:
  Theory and Algorithms. J. Symbolic Computation, Vol. 12 (1991), pp.
  533-583.
"""

#*****************************************************************************
#       Copyright (C) 2006 - 2011 Robert L. Miller <rlmillster@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
#*****************************************************************************

from libc.string cimport memcmp

from sage.groups.perm_gps.partn_ref.data_structures cimport *
from sage.data_structures.bitset_base cimport *
from sage.rings.integer cimport Integer
from sage.matrix.constructor import Matrix
from sage.groups.perm_gps.partn_ref.refinement_binary cimport NonlinearBinaryCodeStruct, refine_by_bip_degree
from sage.groups.perm_gps.partn_ref.double_coset cimport double_coset


cdef class MatrixStruct:

    def __cinit__(self, matrix):
        cdef int i, j
        cdef int *num_rows
        self.degree = matrix.ncols()
        self.nwords = matrix.nrows()
        cdef NonlinearBinaryCodeStruct S_temp
        self.matrix = matrix

        symbols = set(self.matrix.list())
        symbols.discard(0)
        self.symbols = sorted(symbols)
        self.nsymbols = len(symbols)

        self.symbol_structs = []
        num_rows = <int *> sig_malloc(self.nsymbols * sizeof(int))
        self.temp_col_ps = PS_new(self.degree, 1)
        if num_rows is NULL or self.temp_col_ps is NULL:
            sig_free(num_rows)
            PS_dealloc(self.temp_col_ps)
            raise MemoryError

        for i in range(self.nsymbols):
            num_rows[i] = 0
        for row in self.matrix.rows():
            row = set(row.list())
            if 0 in row:
                row.remove(0)
            for s in row:
                num_rows[self.symbols.index(s)] += 1
        for i in range(self.nsymbols):
            S_temp = NonlinearBinaryCodeStruct( (self.degree, num_rows[i]) )
            self.symbol_structs.append(S_temp)

        for i in range(self.nsymbols):
            num_rows[i] = 0
        for row in self.matrix.rows():
            row_list = row.list()
            row_list.reverse()
            for i in row.nonzero_positions():
                s = row[i]
                j = self.symbols.index(s)
                S_temp = <NonlinearBinaryCodeStruct>self.symbol_structs[j]
                bitset_set( &S_temp.words[num_rows[j]], i)
                if row_list.count(s) == 1 or row_list.index(s) == self.degree - i - 1:
                    num_rows[j] += 1
        sig_free(num_rows)
        self.output = NULL

    def __dealloc__(self):
        PS_dealloc(self.temp_col_ps)
        if self.output is not NULL:
            deallocate_agcl_output(self.output)

    def display(self):
        """
        Display the matrix, and associated data.

        EXAMPLES::

            sage: from sage.groups.perm_gps.partn_ref.refinement_matrices import MatrixStruct
            sage: M = MatrixStruct(Matrix(GF(5), [[0,1,1,4,4],[0,4,4,1,1]]))
            sage: M.display()
            [0 1 1 4 4]
            [0 4 4 1 1]
            <BLANKLINE>
            01100
            00011
            1
            <BLANKLINE>
            00011
            01100
            4
        """
        print(self.matrix)
        print("")
        cdef int i,j=0
        cdef NonlinearBinaryCodeStruct S_temp
        for S in self.symbol_structs:
            S_temp = <NonlinearBinaryCodeStruct>S
            for i in range(S_temp.nwords):
                print(bitset_string(&S_temp.words[i]))
            print(self.symbols[j])
            print("")
            j += 1

    def run(self, partition=None):
        """
        Perform the canonical labeling and automorphism group computation,
        storing results to ``self``.

        INPUT:

        - ``partition`` -- an optional list of lists partition of the columns;
          default is the unit partition

        EXAMPLES::

            sage: from sage.groups.perm_gps.partn_ref.refinement_matrices import MatrixStruct

            sage: M = MatrixStruct(matrix(GF(3),[[0,1,2],[0,2,1]]))
            sage: M.run()
            sage: M.automorphism_group()
            ([[0, 2, 1]], 2, [1])
            sage: M.canonical_relabeling()
            [0, 1, 2]

            sage: M = MatrixStruct(matrix(GF(3),[[0,1,2],[0,2,1],[1,0,2],[1,2,0],[2,0,1],[2,1,0]]))
            sage: M.automorphism_group()[1] == 6
            True

            sage: M = MatrixStruct(matrix(GF(3),[[0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,2]]))
            sage: M.automorphism_group()[1] == factorial(14)
            True
        """
        cdef int i, n = self.degree
        cdef PartitionStack *part
        cdef NonlinearBinaryCodeStruct S_temp
        for i in range(self.nsymbols):
            S_temp = <NonlinearBinaryCodeStruct> self.symbol_structs[i]
            S_temp.first_time = 1

        if partition is None:
            part = PS_new(n, 1)
        else:
            part = PS_from_list(partition)
        if part is NULL:
            raise MemoryError

        self.output = get_aut_gp_and_can_lab(<void *> self, part, self.degree, &all_matrix_children_are_equivalent, &refine_matrix, &compare_matrices, 1, NULL, NULL, NULL)

        PS_dealloc(part)

    def automorphism_group(self):
        """
        Return a list of generators of the automorphism group, along with its
        order and a base for which the list of generators is a strong generating
        set.

        For more examples, see self.run().

        EXAMPLES::

            sage: from sage.groups.perm_gps.partn_ref.refinement_matrices import MatrixStruct

            sage: M = MatrixStruct(matrix(GF(3),[[0,1,2],[0,2,1]]))
            sage: M.automorphism_group()
            ([[0, 2, 1]], 2, [1])
        """
        cdef int i, j
        cdef list generators, base
        cdef Integer order
        if self.output is NULL:
            self.run()
        generators = [[self.output.generators[i * self.degree + j]
                       for j in range(self.degree)]
                      for i in range(self.output.num_gens)]
        order = Integer()
        SC_order(self.output.group, 0, order.value)
        base = [self.output.group.base_orbits[i][0] for i in range(self.output.group.base_size)]
        return generators, order, base

    def canonical_relabeling(self):
        """
        Return a canonical relabeling (in list permutation format).

        For more examples, see self.run().

        EXAMPLES::

            sage: from sage.groups.perm_gps.partn_ref.refinement_matrices import MatrixStruct

            sage: M = MatrixStruct(matrix(GF(3),[[0,1,2],[0,2,1]]))
            sage: M.canonical_relabeling()
            [0, 1, 2]
        """
        cdef int i
        if self.output is NULL:
            self.run()
        return [self.output.relabeling[i] for i in range(self.degree)]

    def is_isomorphic(self, MatrixStruct other):
        """
        Calculate whether ``self`` is isomorphic to ``other``.

        EXAMPLES::

            sage: from sage.groups.perm_gps.partn_ref.refinement_matrices import MatrixStruct
            sage: M = MatrixStruct(Matrix(GF(11), [[1,2,3,0,0,0],[0,0,0,1,2,3]]))
            sage: N = MatrixStruct(Matrix(GF(11), [[0,1,0,2,0,3],[1,0,2,0,3,0]]))
            sage: M.is_isomorphic(N)
            [0, 2, 4, 1, 3, 5]
        """
        cdef int i, n = self.degree
        cdef int *output
        cdef int *ordering
        cdef PartitionStack *part
        cdef NonlinearBinaryCodeStruct S_temp
        for i in range(self.nsymbols):
            S_temp = self.symbol_structs[i]
            S_temp.first_time = 1
            S_temp = other.symbol_structs[i]
            S_temp.first_time = 1
        part = PS_new(n, 1)
        ordering = <int *> sig_malloc(self.degree * sizeof(int))
        output = <int *> sig_malloc(self.degree * sizeof(int))
        if part is NULL or ordering is NULL or output is NULL:
            PS_dealloc(part)
            sig_free(ordering)
            sig_free(output)
            raise MemoryError
        for i in range(self.degree):
            ordering[i] = i

        cdef bint isomorphic = double_coset(<void *> self, <void *> other, part, ordering, self.degree, &all_matrix_children_are_equivalent, &refine_matrix, &compare_matrices, NULL, NULL, output)

        PS_dealloc(part)
        sig_free(ordering)
        if isomorphic:
            output_py = [output[i] for i in range(self.degree)]
        else:
            output_py = False
        sig_free(output)
        return output_py

cdef int refine_matrix(PartitionStack *PS, void *S, int *cells_to_refine_by, int ctrb_len) noexcept:
    cdef MatrixStruct M = <MatrixStruct> S
    cdef int temp_inv, invariant = 1
    cdef bint changed = 1
    while changed:
        PS_copy_from_to(PS, M.temp_col_ps)
        for BCS in M.symbol_structs:
            temp_inv = refine_by_bip_degree(PS, <void *> BCS, cells_to_refine_by, ctrb_len)
            invariant *= temp_inv + 1
        if memcmp(PS.entries, M.temp_col_ps.entries, 2*M.degree * sizeof(int)) == 0:
            changed = 0
    return invariant

cdef int compare_matrices(int *gamma_1, int *gamma_2, void *S1, void *S2, int degree) noexcept:
    cdef MatrixStruct MS1 = <MatrixStruct> S1
    cdef MatrixStruct MS2 = <MatrixStruct> S2
    M1 = MS1.matrix
    M2 = MS2.matrix
    cdef int i
    MM1 = Matrix(M1.base_ring(), M1.nrows(), M1.ncols(), sparse=M1.is_sparse())
    MM2 = Matrix(M2.base_ring(), M2.nrows(), M2.ncols(), sparse=M2.is_sparse())
    for i in range(degree):
        MM1.set_column(i, M1.column(gamma_1[i]))
        MM2.set_column(i, M2.column(gamma_2[i]))
    rows1 = sorted(MM1.rows())
    rows2 = sorted(MM2.rows())
    if rows1 == rows2:
        return 0
    return -1 if rows1 < rows2 else 1

cdef bint all_matrix_children_are_equivalent(PartitionStack *PS, void *S) noexcept:
    return 0


def random_tests(n=10, nrows_max=50, ncols_max=50, nsymbols_max=10, perms_per_matrix=5, density_range=(.1,.9)):
    """
    Test to make sure that ``C(gamma(M)) == C(M)`` for random permutations ``gamma``
    and random matrices ``M``, and that ``M.is_isomorphic(gamma(M))`` returns an
    isomorphism.

    INPUT:

    - ``n`` -- run tests on this many matrices
    - ``nrows_max`` -- test matrices with at most this many rows
    - ``ncols_max`` -- test matrices with at most this many columns
    - ``perms_per_matrix`` -- test each matrix with this many random permutations
    - ``nsymbols_max`` -- maximum number of distinct symbols in the matrix

    This code generates n random matrices M on at most ncols_max columns and at
    most nrows_max rows. The density of entries in the basis is chosen randomly
    between 0 and 1.

    For each matrix M generated, we uniformly generate perms_per_matrix random
    permutations and verify that the canonical labels of M and the image of M
    under the generated permutation are equal, and that the isomorphism is
    discovered by the double coset function.

    TESTS::

        sage: import sage.groups.perm_gps.partn_ref.refinement_matrices
        sage: sage.groups.perm_gps.partn_ref.refinement_matrices.random_tests()  # long time (up to 30s on sage.math, 2011)
        All passed: ... random tests on ... matrices.
    """
    from sage.misc.prandom import random, randint
    from sage.combinat.permutation import Permutations
    from sage.matrix.constructor import random_matrix, matrix
    from sage.rings.finite_rings.finite_field_constructor import FiniteField as GF
    from sage.arith.misc import next_prime
    cdef int i, j, nrows, num_tests = 0, num_matrices = 0
    cdef MatrixStruct M, N
    for m in range(n):
        p = random()*(density_range[1]-density_range[0]) + density_range[0]
        nrows = randint(1, nrows_max)
        ncols = randint(1, ncols_max)
        nsymbols = next_prime(randint(1, nsymbols_max))
        S = Permutations(ncols)
        MM = random_matrix(GF(nsymbols), nrows, ncols, sparse=False, density=p)
        M = MatrixStruct( MM )
        M.run()

        for i in range(perms_per_matrix):
            perm = [a-1 for a in list(S.random_element())]
            NN = matrix(GF(nsymbols), nrows, ncols)
            for j in range(ncols):
                NN.set_column(perm[j], MM.column(j))
            N = MatrixStruct(NN)
            # now N is a random permutation of M
            N.run()

            M_relab = M.canonical_relabeling()
            N_relab = N.canonical_relabeling()

            M_C = matrix(GF(nsymbols), nrows, ncols)
            N_C = matrix(GF(nsymbols), nrows, ncols)

            for j in range(ncols):
                M_C.set_column(M_relab[j], MM.column(j))
                N_C.set_column(N_relab[j], NN.column(j))

            M_C = matrix(GF(nsymbols), sorted(M_C.rows()))
            N_C = matrix(GF(nsymbols), sorted(N_C.rows()))

            if M_C != N_C:
                print("M:")
                print(M.matrix.str())
                print("perm:")
                print(perm)
                return

            isom = M.is_isomorphic(N)
            if not isom:
                print("isom FAILURE: M:")
                print(M.matrix.str())
                print("isom FAILURE: N:")
                print(N.matrix.str())
                return

        num_tests += perms_per_matrix
        num_matrices += 2
    print("All passed: %d random tests on %d matrices." %
          (num_tests, num_matrices))



# sage_setup: distribution = sagemath-flint
r"""
Linkage for arithmetic with FLINT's nmod_poly_t elements.

This file provides the backend for \class{Polynomial_zmod_flint} via
templating.

AUTHOR:

- Martin Albrecht (2009-01) another initial implementation
- Burcin Erocal (2008-11) initial implementation
"""
#*****************************************************************************
#       Copyright (C) 2008-2009 Burcin Erocal <burcin@erocal.org>
#       Copyright (C) 2009 Martin Albrecht <M.R.Albrecht@rhul.ac.uk>
#
#  Distributed under the terms of the GNU General Public License (GPL),
#  version 2 or any later version.  The full text of the GPL is available at:
#                  https://www.gnu.org/licenses/
#*****************************************************************************

from cysignals.signals cimport sig_on, sig_off
from cysignals.memory cimport sig_malloc, sig_free

from sage.libs.flint.types cimport *
from sage.libs.flint.nmod_poly cimport *
from sage.libs.flint.nmod_poly_factor cimport *
from sage.libs.flint.ulong_extras cimport *
from sage.structure.factorization import Factorization

cdef inline celement *celement_new(unsigned long n) noexcept:
    cdef celement *g = <celement *>sig_malloc(sizeof(nmod_poly_t))
    nmod_poly_init(g, n)
    return g

cdef inline int celement_delete(nmod_poly_t e, unsigned long n) noexcept:
    nmod_poly_clear(e)
    sig_free(e)

cdef inline int celement_construct(nmod_poly_t e, unsigned long n) noexcept:
    """
    EXAMPLES::

        sage: P.<x> = GF(32003)[]                                                       # needs sage.rings.finite_rings

        sage: Q.<x> = GF(7)[]
    """
    nmod_poly_init(e, n)

cdef inline int celement_destruct(nmod_poly_t e, unsigned long n) noexcept:
    """
    EXAMPLES::

        sage: P.<x> = GF(32003)[]                                                       # needs sage.rings.finite_rings
        sage: del x                                                                     # needs sage.rings.finite_rings

        sage: Q.<x> = GF(7)[]
        sage: del x
    """
    nmod_poly_clear(e)

cdef inline int celement_gen(nmod_poly_t e, long i, unsigned long n) except -2:
    """
    EXAMPLES::

        sage: P.<x> = GF(32003)[]                                                       # needs sage.rings.finite_rings

        sage: Q.<x> = GF(7)[]
    """
    nmod_poly_zero(e)
    nmod_poly_set_coeff_ui(e, 1, 1)

cdef object celement_repr(nmod_poly_t e, unsigned long n):
    raise NotImplementedError

cdef inline int celement_set(nmod_poly_t res, nmod_poly_t a, unsigned long n) except -2:
    """
    EXAMPLES::

        sage: # needs sage.rings.finite_rings
        sage: P.<x> = GF(32003)[]
        sage: y = copy(x)
        sage: y is x
        False
        sage: y == x
        True

        sage: Q.<x> = GF(7)[]
        sage: y = copy(x)
        sage: y is x
        False
        sage: y == x
        True

        sage: R.<x> = PolynomialRing(Integers(121))
        sage: S.<y> = PolynomialRing(Integers(11))
        sage: S(50*x)
        6*y
        sage: R(S(50*x))
        6*x
    """
    cdef unsigned long i
    if a.mod.n <= n:
        nmod_poly_set(res, a)
    else:
        nmod_poly_zero(res)
        for i from 0 <= i < a.length:
            nmod_poly_set_coeff_ui(res, i, nmod_poly_get_coeff_ui(a, i) % n)

cdef inline int celement_set_si(nmod_poly_t res, long i, unsigned long n) except -2:
    """
    EXAMPLES::

        sage: # needs sage.rings.finite_rings
        sage: P.<x> = GF(32003)[]
        sage: P(32003)
        0
        sage: P(1)
        1
        sage: P(32004)
        1

        sage: Q.<x> = GF(7)[]
        sage: Q(7)
        0
        sage: Q(1)
        1
        sage: Q(8)
        1
    """
    while i < 0:
        i += n
    nmod_poly_zero(res)
    if i:
        nmod_poly_set_coeff_ui(res, 0, <unsigned long>i)

cdef inline long celement_get_si(nmod_poly_t res, unsigned long n) except -2:
    raise NotImplementedError

cdef inline bint celement_is_zero(nmod_poly_t a, unsigned long n) except -2:
    """
    EXAMPLES::

        sage: # needs sage.rings.finite_rings
        sage: P.<x> = GF(32003)[]
        sage: P(1).is_zero()
        False
        sage: P(0).is_zero()
        True

        sage: Q.<x> = GF(7)[]
        sage: Q(1).is_zero()
        False
        sage: Q(0).is_zero()
        True
    """
    return nmod_poly_is_zero(a)

cdef inline bint celement_is_one(nmod_poly_t a, unsigned long n) except -2:
    """
    EXAMPLES::

        sage: # needs sage.rings.finite_rings
        sage: P.<x> = GF(32003)[]
        sage: P(1).is_one()
        True
        sage: P(0).is_one()
        False

        sage: Q.<x> = GF(7)[]
        sage: Q(1).is_one()
        True
        sage: Q(0).is_one()
        False
    """

    return nmod_poly_is_one(a)

cdef inline bint celement_equal(nmod_poly_t a, nmod_poly_t b, unsigned long n) except -2:
    """
    EXAMPLES::

        sage: # needs sage.rings.finite_rings
        sage: P.<x> = GF(32003)[]
        sage: (3*2)*x == 3*(2*x)
        True
        sage: (3*2)*x + 2 == 3*(2*x) + 1 + 1
        True
        sage: (3*2)*x + 7 == 3*(2*x) + 1 + 1
        False

        sage: Q.<x> = GF(7)[]
        sage: (3*2)*x == 3*(2*x)
        True
        sage: (3*2)*x + 2 == 3*(2*x) + 1 + 1
        True
        sage: (3*2)*x + 7 == 3*(2*x) + 1 + 1
        False
    """
    return nmod_poly_equal(a, b)

cdef inline int celement_cmp(nmod_poly_t l, nmod_poly_t r, unsigned long n) except -2:
    """
    EXAMPLES::

        sage: # needs sage.rings.finite_rings
        sage: P.<x> = GF(32003)[]
        sage: x > x
        False
        sage: x^2 > x
        True
        sage: 3*x > x
        True

        sage: Q.<x> = GF(7)[]
        sage: x > x
        False
        sage: x^2 > x
        True
        sage: 3*x > x
        True

        sage: f = x^64 + x^20 + 1
        sage: g = x^63 + x^20 + 1
        sage: f > g
        True

        sage: f = x^64 + x^10 + 1
        sage: g = x^64 + x^20 + 1
        sage: f < g
        True

        sage: f = x^64 + x^20
        sage: g = x^64 + x^20 + 1
        sage: f < g
        True
    """
    cdef int deg_right = nmod_poly_degree(r)
    cdef int degdiff = deg_right - nmod_poly_degree(l)
    cdef int i
    cdef unsigned long rcoeff, lcoeff
    if degdiff > 0:
        return -1
    elif degdiff < 0:
        return 1
    else:
        if nmod_poly_equal(l, r):
            return 0
        i = deg_right
        while i >= 0:
            rcoeff = nmod_poly_get_coeff_ui(r, i)
            lcoeff = nmod_poly_get_coeff_ui(l, i)
            if lcoeff < rcoeff:
                return -1
            if lcoeff > rcoeff:
                return 1
            i -= 1
        return 0

cdef long celement_len(nmod_poly_t a, unsigned long n) except -2:
    """
    EXAMPLES::

        sage: # needs sage.rings.finite_rings
        sage: P.<x> = GF(32003)[]
        sage: (x + 1).degree()
        1
        sage: (x).degree()
        1
        sage: P(0).degree()
        -1

        sage: Q.<x> = GF(7)[]
        sage: (x + 1).degree()
        1
        sage: (x).degree()
        1
        sage: Q(0).degree()
        -1
    """
    return <long>nmod_poly_length(a)

cdef inline int celement_add(nmod_poly_t res, nmod_poly_t a, nmod_poly_t b, unsigned long n) except -2:
    """
    EXAMPLES::

        sage: # needs sage.rings.finite_rings
        sage: P.<x> = GF(32003)[]
        sage: x + 1
        x + 1

        sage: Q.<x> = GF(7)[]
        sage: x + 1
        x + 1
    """
    nmod_poly_add(res, a, b)

cdef inline int celement_sub(nmod_poly_t res, nmod_poly_t a, nmod_poly_t b, unsigned long n) except -2:
    """
    EXAMPLES::

        sage: # needs sage.rings.finite_rings
        sage: P.<x> = GF(32003)[]
        sage: x - 1
        x + 32002

        sage: Q.<x> = GF(7)[]
        sage: x - 1
        x + 6
    """
    nmod_poly_sub(res, a, b)

cdef inline int celement_neg(nmod_poly_t res, nmod_poly_t a, unsigned long n) except -2:
    """
    EXAMPLES::

        sage: # needs sage.rings.finite_rings
        sage: P.<x> = GF(32003)[]
        sage: -(x + 2)
        32002*x + 32001

        sage: Q.<x> = GF(7)[]
        sage: -(x + 2)
        6*x + 5
    """
    nmod_poly_neg(res, a)

cdef inline int celement_mul_scalar(nmod_poly_t res, nmod_poly_t p,
        object c, unsigned long n) except -2:
    """
    TESTS::

        sage: # needs sage.rings.finite_rings
        sage: P.<x> = GF(32003)[]
        sage: p = P.random_element(degree=2)
        sage: (389*p).coefficients() == [389*x for x in p.coefficients()]
        True
        sage: p = P.random_element(degree=8)
        sage: (p*9836).coefficients() == [x*9836 for x in p.coefficients()]
        True
    """
    nmod_poly_scalar_mul_nmod(res, p, (<unsigned long>c) % n)

cdef inline int celement_mul(nmod_poly_t res, nmod_poly_t a, nmod_poly_t b, unsigned long n) except -2:
    """
    EXAMPLES::

        sage: # needs sage.rings.finite_rings
        sage: P.<x> = GF(32003)[]
        sage: (x + 1) * (x + 2)
        x^2 + 3*x + 2

        sage: Q.<x> = GF(7)[]
        sage: (x + 1) * (x + 2)
        x^2 + 3*x + 2
    """
    nmod_poly_mul(res, a, b)

cdef inline int celement_div(nmod_poly_t res, nmod_poly_t a, nmod_poly_t b, unsigned long n) except -2:
    raise NotImplementedError

cdef inline int celement_truncate(nmod_poly_t res, nmod_poly_t a, long len, unsigned long n) except -2:
    """
    EXAMPLES::

        sage: P.<x> = GF(7)[]
        sage: p = 4*x^4 + 3*x^3 + 2*x^2 + x
        sage: p.truncate(3)
        2*x^2 + x

        sage: # needs sage.rings.finite_rings
        sage: Q.<x> = GF(32003)[]
        sage: q = 1 + x + x^2 * Q.random_element()
        sage: q.truncate(2)
        x + 1
    """
    nmod_poly_set(res, a)
    nmod_poly_truncate(res, len)

cdef inline int celement_floordiv(nmod_poly_t res, nmod_poly_t a, nmod_poly_t b, unsigned long n) except -2:
    """
    EXAMPLES::

        sage: # needs sage.rings.finite_rings
        sage: P.<x> = GF(32003)[]
        sage: (x + 1) // (x + 2)
        1
        sage: (3*x^2 + 1) // (x + 2)
        3*x + 31997
        sage: (x^2 + 3*x + 2)//(x + 1)
        x + 2
        sage: (x^2 + 3*x + 2)//(x + 2)
        x + 1

        sage: Q.<x> = GF(7)[]
        sage: (x + 1) // (x + 2)
        1
        sage: (3*x^2 + 1) // (x + 2)
        3*x + 1
        sage: (x^2 + 3*x + 2)//(x + 1)
        x + 2
        sage: (x^2 + 3*x + 2)//(x + 2)
        x + 1
    """
    nmod_poly_div(res, a, b)

cdef inline int celement_mod(nmod_poly_t res, nmod_poly_t a, nmod_poly_t b, unsigned long n) except -2:
    """
    EXAMPLES::

        sage: # needs sage.rings.finite_rings
        sage: P.<x> = GF(32003)[]
        sage: f = 24998*x^2 + 29761*x + 2252
        sage: g = 20778*x^2 + 15346*x + 12697
        sage: f % g
        5815*x + 10280
        sage: f^5 % g
        7231*x + 17274

        sage: R.<x> = Integers(81)[]
        sage: f = x^7 + x + 1; g = x^3
        sage: r = f % g; r
        x + 1
        sage: g * x^4 + r
        x^7 + x + 1
        sage: f = x^3 + 1
        sage: f % 3
        Traceback (most recent call last):
        ...
        ValueError: Leading coefficient of a must be invertible.
        sage: f % 0
        Traceback (most recent call last):
        ...
        ZeroDivisionError
    """
    cdef nmod_poly_t q
    cdef unsigned long leadcoeff, modulus

    nmod_poly_init(q, n)
    leadcoeff = nmod_poly_get_coeff_ui(b, nmod_poly_degree(b))
    modulus = nmod_poly_modulus(b)
    if (leadcoeff > 1 and n_gcd(modulus,leadcoeff) != 1):
        raise ValueError("Leading coefficient of a must be invertible.")

    nmod_poly_divrem(q, res, a, b)
    nmod_poly_clear(q)

cdef inline int celement_quorem(nmod_poly_t q, nmod_poly_t r, nmod_poly_t a, nmod_poly_t b, unsigned long n) except -2:
    """
    EXAMPLES::

        sage: R.<x> = Integers(125)[]
        sage: f = x^5+1; g = (x+1)^2
        sage: q, r = f.quo_rem(g)
        sage: q
        x^3 + 123*x^2 + 3*x + 121
        sage: r
        5*x + 5
        sage: q*g + r
        x^5 + 1

        sage: x.quo_rem(5*x)
        Traceback (most recent call last):
        ...
        ValueError: Leading coefficient of a must be invertible.

        sage: x.quo_rem(0)
        Traceback (most recent call last):
        ...
        ZeroDivisionError
    """
    cdef unsigned long leadcoeff, modulus

    leadcoeff = nmod_poly_get_coeff_ui(b, nmod_poly_degree(b))
    modulus = nmod_poly_modulus(b)
    if (leadcoeff > 1 and n_gcd(modulus,leadcoeff) != 1):
        raise ValueError("Leading coefficient of a must be invertible.")

    nmod_poly_divrem(q, r, a, b)

cdef inline int celement_inv(nmod_poly_t res, nmod_poly_t a, unsigned long n) except -2:
    raise NotImplementedError

cdef inline int celement_pow(nmod_poly_t res, nmod_poly_t x, long e, nmod_poly_t modulus, unsigned long n) except -2:
    """
    Compute `x^e`, possibly modulo ``modulus``.

    INPUT:

    - ``x`` -- polynomial; the base

    - ``e`` -- integer; the exponent

    - ``modulus`` -- polynomial or NULL; if not NULL, then perform a modular exponentiation

    - ``n`` -- integer; not used, but all polynomials' coefficients are understood modulo ``n``

    EXAMPLES::

        sage: # needs sage.rings.finite_rings
        sage: P.<x> = GF(32003)[]
        sage: f = 24998*x^2 + 29761*x + 2252
        sage: f*f
        9426*x^4 + 15477*x^3 + 6531*x^2 + 14980*x + 15030
        sage: f^2
        9426*x^4 + 15477*x^3 + 6531*x^2 + 14980*x + 15030
        sage: f*f*f
        25062*x^6 + 30670*x^5 + 15816*x^4 + 20746*x^3 + 9142*x^2 + 5697*x + 20389
        sage: f^3
        25062*x^6 + 30670*x^5 + 15816*x^4 + 20746*x^3 + 9142*x^2 + 5697*x + 20389
        sage: f*f*f*f*f
        20269*x^10 + 20535*x^9 + 7313*x^8 + 7311*x^7 + 16853*x^6 + 142*x^5 + 23853*x^4 + 12065*x^3 + 516*x^2 + 8473*x + 17945
        sage: f^5
        20269*x^10 + 20535*x^9 + 7313*x^8 + 7311*x^7 + 16853*x^6 + 142*x^5 + 23853*x^4 + 12065*x^3 + 516*x^2 + 8473*x + 17945
        sage: f^0
        1
        sage: f^1
        24998*x^2 + 29761*x + 2252
        sage: f^-1
        18649/(x^2 + 16863*x + 9612)
        sage: f^-5
        24620/(x^10 + 20309*x^9 + 29185*x^8 + 11948*x^7 + 1965*x^6 + 7713*x^5 + 5810*x^4 + 20457*x^3 + 30732*x^2 + 9706*x + 4485)

     Testing the modulus::

        sage: # needs sage.rings.finite_rings
        sage: g = 20778*x^2 + 15346*x + 12697
        sage: pow(f, 2, g)
        15328*x + 6968
        sage: f^2 % g
        15328*x + 6968
        sage: pow(f, -2, g)
        16346/(x + 251)
        sage: (f^2 % g)^-1
        16346/(x + 251)
        sage: pow(f, 5, g)
        7231*x + 17274
        sage: f^5 % g
        7231*x + 17274

    Make sure that exponentiation can be interrupted, see :issue:`17470`::

        sage: # needs sage.rings.finite_rings
        sage: n = 2^23
        sage: from sage.doctest.util import ensure_interruptible_after
        sage: with ensure_interruptible_after(0.2): (x^n).degree()
    """
    if modulus != NULL:
        sig_on()
        nmod_poly_powmod_ui_binexp(res, x, e, modulus)
        sig_off()
    else:
        sig_on()
        nmod_poly_pow(res, x, e)
        sig_off()

cdef inline int celement_gcd(nmod_poly_t res, nmod_poly_t a, nmod_poly_t b, unsigned long n) except -2:
    """
    EXAMPLES::

        sage: # needs sage.rings.finite_rings
        sage: P.<x> = GF(32003)[]
        sage: f = P.random_element(degree=4)
        sage: g = P.random_element(degree=3)
        sage: h = P.random_element(degree=2)
        sage: F = f*g
        sage: G = f*h
        sage: d = (F).gcd(G)
        sage: (F//d)*d == F
        True
        sage: (G//d)*d == G
        True

        sage: Q.<x> = GF(7)[]
        sage: f = Q.random_element(degree=4)
        sage: g = Q.random_element(degree=3)
        sage: h = Q.random_element(degree=2)
        sage: F = f*g
        sage: G = f*h
        sage: d = (F).gcd(G)
        sage: (F//d)*d == F
        True
        sage: (G//d)*d == G
        True

    Check that we catch a case where FLINT fails.  These generate the unit
    ideal, so their GCD should be 1 (or another unit)::

        sage: R.<x> = Integers(121)[]
        sage: f = 11*x^2 + 1
        sage: f - 61*x*f.derivative()
        1
        sage: gcd(f, f.derivative())
        Traceback (most recent call last):
        ...
        RuntimeError: FLINT gcd calculation failed
    """
    if celement_is_zero(b, n):
        nmod_poly_set(res, a)
        return 0

    # FLINT provides no interface for detecting errors here
    try:
        sig_on()
        try:
            nmod_poly_gcd(res, a, b)
        finally:
            sig_off()
    except RuntimeError:
        raise RuntimeError("FLINT gcd calculation failed")

    cdef unsigned long leadcoeff = nmod_poly_get_coeff_ui(res, nmod_poly_degree(res))
    cdef unsigned long modulus = nmod_poly_modulus(res)
    if n_gcd(modulus, leadcoeff) == 1:
        nmod_poly_make_monic(res, res)

cdef inline int celement_xgcd(nmod_poly_t res, nmod_poly_t s, nmod_poly_t t, nmod_poly_t a, nmod_poly_t b, unsigned long n) except -2:
    """
    EXAMPLES::

        sage: # needs sage.rings.finite_rings
        sage: P.<x> = GF(32003)[]
        sage: f = P.random_element(degree=4)
        sage: g = P.random_element(degree=3)
        sage: h = P.random_element(degree=2)
        sage: F = f*g
        sage: G = f*h
        sage: d,s,t = (F).xgcd(G)
        sage: (F//d)*d == F
        True
        sage: (G//d)*d == G
        True

        sage: Q.<x> = GF(7)[]
        sage: f = Q.random_element(degree=4)
        sage: g = Q.random_element(degree=3)
        sage: h = Q.random_element(degree=2)
        sage: F = f*g
        sage: G = f*h
        sage: d,s,t = (F).xgcd(G)
        sage: (F//d)*d == F
        True
        sage: (G//d)*d == G
        True

    TESTS:

    Ensure that :issue:`38537` is fixed::

        sage: k = Zmod(2**16)
        sage: R.<x> = k[]
        sage: u = x + 10161
        sage: v = x + 10681
        sage: u.xgcd(v)
        Traceback (most recent call last):
        ...
        ValueError: non-invertible elements encountered during XGCD
    """
    try:
        sig_on()
        nmod_poly_xgcd(res, s, t, a, b)
        sig_off()
    except RuntimeError:
        raise ValueError("non-invertible elements encountered during XGCD")


cdef factor_helper(Polynomial_zmod_flint poly, bint squarefree=False):
    """
    EXAMPLES::

        sage: # needs sage.rings.finite_rings
        sage: P.<x> = GF(1009)[]
        sage: factors = (prod(P.random_element(degree=2) for i in range(5))).factor()
        sage: all(factor.is_irreducible() for factor, mult in factors)
        True
        sage: def nonzero_random(P):
        ....:     r = P.random_element()
        ....:     while r == 0:
        ....:         r = P.random_element()
        ....:     return r
        sage: p = (prod(nonzero_random(P)^i for i in range(5)))
        sage: decomp = p.squarefree_decomposition()
        sage: any(factor.is_square() for factor, mult in decomp)
        False
        sage: start = 0
        sage: for factor, mult in decomp:
        ....:     assert mult > start
        ....:     start = mult
    """
    cdef nmod_poly_factor_t factors_c
    nmod_poly_factor_init(factors_c)

    sig_on()
    if squarefree:
        nmod_poly_factor_squarefree(factors_c, &poly.x)
    else:
        nmod_poly_factor(factors_c, &poly.x)
    sig_off()

    factor_list = []
    cdef Polynomial_zmod_flint t
    for i in range(factors_c.num):
        t = poly._new()
        nmod_poly_swap(&t.x, &factors_c.p[i])
        factor_list.append((t, factors_c.exp[i]))

    nmod_poly_factor_clear(factors_c)

    return Factorization(factor_list, unit=poly.leading_coefficient(),
            sort=(not squarefree))



# sage_setup: distribution = sagemath-modules
from sage.data_structures.bitset cimport bitset_t
from sage.matroids.basis_exchange_matroid cimport BasisExchangeMatroid
from sage.matroids.set_system cimport SetSystem

cdef class BasisMatroid(BasisExchangeMatroid):
    cdef bitset_t _bb
    cdef bitset_t _b
    cdef SetSystem _nonbases
    cdef Py_hash_t _bases_invariant_var
    cdef SetSystem _bases_partition_var
    cdef Py_hash_t _bases_invariant2_var
    cdef SetSystem _bases_partition2_var
    cdef Py_hash_t _bases_invariant3_var
    cdef SetSystem _bases_partition3_var

    cdef reset_current_basis(self)

    cpdef bint _is_basis(self, frozenset X) noexcept

    cpdef bases_count(self)
    cpdef SetSystem bases(self)
    cpdef SetSystem nonbases(self)

    cpdef truncation(self)
    cpdef _extension(self, e, H)
    cpdef _with_coloop(self, e)
    # cpdef relabel(self, mapping)

    cpdef Py_hash_t _bases_invariant(self) noexcept
    cpdef SetSystem _bases_partition(self)
    cpdef Py_hash_t _bases_invariant2(self) noexcept
    cpdef SetSystem _bases_partition2(self)
    cpdef Py_hash_t _bases_invariant3(self) noexcept
    cpdef SetSystem _bases_partition3(self)
    cdef _reset_invariants(self)
    cpdef  bint is_distinguished(self, e) noexcept
    cpdef _is_relaxation(self, M, morphism)
    cpdef _is_isomorphism(self, M, morphism)
    cpdef _isomorphism(self, other)
    cpdef _is_isomorphic(self, other, certificate=*)

cdef  binom_init(long n, long k)
cdef  long set_to_index(bitset_t S) noexcept
cdef  index_to_set(bitset_t, long, long, long)



# sage_setup: distribution = sagemath-modules
"""
Counting Congruence Solutions

This file provides more user-friendly Python front-ends to the Cython code in
:mod:`sage.quadratic_forms.count_local`.
"""
################################################################
# Methods for counting/computing the number of representations #
# of a number by a quadratic form in Z/(p^k)Z of various types #
################################################################

from sage.quadratic_forms.count_local_2 import CountAllLocalTypesNaive


def count_congruence_solutions_as_vector(self, p, k, m, zvec, nzvec):
    r"""
    Return the number of integer solution vectors `x` satisfying the
    congruence `Q(x) = m` (mod `p^k`) of each solution type (i.e. All,
    Good, Zero, Bad, BadI, BadII) which satisfy the additional
    congruence conditions of having certain coefficients = 0 (mod `p`)
    and certain collections of coefficients not congruent to the zero
    vector (mod `p`).

    A solution vector `x` satisfies the additional congruence conditions
    specified by ``zvec`` and ``nzvec`` (and therefore is counted) iff both of
    the following conditions hold:

    1. `x_i = 0` (mod `p`) for all `i` in ``zvec``

    2. `x_i \neq 0` (mod `p`) for all `i` in ``nzvec``


    REFERENCES:

    See Hanke's (????) paper "Local Densities and explicit bounds...", p??? for
    the definitions of the solution types and congruence conditions.

    INPUT:

    - ``p`` -- prime number > 0
    - ``k`` -- integer > 0
    - ``m`` -- integer (depending only on mod `p^k`)
    - ``zvec``, ``nzvec`` -- lists of integers in ``range(self.dim())``, or ``None``

    OUTPUT:

    a list of six integers `\geq 0` representing the solution types:
    [All, Good, Zero, Bad, BadI, BadII]

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,2,3])
        sage: Q.count_congruence_solutions_as_vector(3, 1, 1, [], [])
        [0, 0, 0, 0, 0, 0]
        sage: Q.count_congruence_solutions_as_vector(3, 1, 1, None, [])
        [0, 0, 0, 0, 0, 0]
        sage: Q.count_congruence_solutions_as_vector(3, 1, 1, [], None)
        [6, 6, 0, 0, 0, 0]
        sage: Q.count_congruence_solutions_as_vector(3, 1, 1, None, None)
        [6, 6, 0, 0, 0, 0]
        sage: Q.count_congruence_solutions_as_vector(3, 1, 2, None, None)
        [6, 6, 0, 0, 0, 0]
        sage: Q.count_congruence_solutions_as_vector(3, 1, 0, None, None)
        [15, 12, 1, 2, 0, 2]
    """
    return CountAllLocalTypesNaive(self, p, k, m, zvec, nzvec)


# //////////////////////////////////////////
# // Front-ends for our counting routines //
# //////////////////////////////////////////

def count_congruence_solutions(self, p, k, m, zvec, nzvec):
    r"""
    Count all solutions of `Q(x) = m` (mod `p^k`) satisfying the
    additional congruence conditions described in
    :meth:`QuadraticForm.count_congruence_solutions_as_vector`.

    INPUT:

    - ``p`` -- prime number > 0

    - ``k`` -- integer > 0

    - ``m`` -- integer (depending only on mod `p^k`)

    - ``zvec``, ``nzvec`` -- lists of integers in ``range(self.dim())``, or ``None``

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,2,3])
        sage: Q.count_congruence_solutions(3, 1, 0, None, None)
        15
    """
    return CountAllLocalTypesNaive(self, p, k, m, zvec, nzvec)[0]


def count_congruence_solutions__good_type(self, p, k, m, zvec, nzvec):
    r"""
    Count the good-type solutions of `Q(x) = m` (mod `p^k`) satisfying the
    additional congruence conditions described in
    :meth:`QuadraticForm.count_congruence_solutions_as_vector`.

    INPUT:

    - ``p`` -- prime number > 0

    - ``k`` -- integer > 0

    - ``m`` -- integer (depending only on mod `p^k`)

    - ``zvec``, ``nzvec`` -- lists of integers up to dim(`Q`)

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,2,3])
        sage: Q.count_congruence_solutions__good_type(3, 1, 0, None, None)
        12
    """
    return CountAllLocalTypesNaive(self, p, k, m, zvec, nzvec)[1]


def count_congruence_solutions__zero_type(self, p, k, m, zvec, nzvec):
    """
    Count the zero-type solutions of `Q(x) = m` (mod `p^k`) satisfying the
    additional congruence conditions described in
    :meth:`QuadraticForm.count_congruence_solutions_as_vector`.

    INPUT:

    - ``p`` -- prime number > 0

    - ``k`` -- integer > 0

    - ``m`` -- integer (depending only on mod `p^k`)

    - ``zvec``, ``nzvec`` -- lists of integers up to dim(`Q`)

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,2,3])
        sage: Q.count_congruence_solutions__zero_type(3, 1, 0, None, None)
        1
    """
    return CountAllLocalTypesNaive(self, p, k, m, zvec, nzvec)[2]


def count_congruence_solutions__bad_type(self, p, k, m, zvec, nzvec):
    """
    Count the bad-type solutions of `Q(x) = m` (mod `p^k`) satisfying the
    additional congruence conditions described in
    :meth:`QuadraticForm.count_congruence_solutions_as_vector`.

    INPUT:

    - ``p`` -- prime number > 0

    - ``k`` -- integer > 0

    - ``m`` -- integer (depending only on mod `p^k`)

    - ``zvec``, ``nzvec`` -- lists of integers up to dim(`Q`)

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,2,3])
        sage: Q.count_congruence_solutions__bad_type(3, 1, 0, None, None)
        2
    """
    return CountAllLocalTypesNaive(self, p, k, m, zvec, nzvec)[3]


def count_congruence_solutions__bad_type_I(self, p, k, m, zvec, nzvec):
    """
    Count the bad-typeI solutions of `Q(x) = m` (mod `p^k`) satisfying
    the additional congruence conditions described in
    :meth:`QuadraticForm.count_congruence_solutions_as_vector`.

    INPUT:

    - ``p`` -- prime number > 0

    - ``k`` -- integer > 0

    - ``m`` -- integer (depending only on mod `p^k`)

    - ``zvec``, ``nzvec`` -- lists of integers up to dim(`Q`)

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,2,3])
        sage: Q.count_congruence_solutions__bad_type_I(3, 1, 0, None, None)
        0
    """
    return CountAllLocalTypesNaive(self, p, k, m, zvec, nzvec)[4]


def count_congruence_solutions__bad_type_II(self, p, k, m, zvec, nzvec):
    """
    Count the bad-typeII solutions of `Q(x) = m` (mod `p^k`) satisfying
    the additional congruence conditions described in
    :meth:`QuadraticForm.count_congruence_solutions_as_vector`.

    INPUT:

    - ``p`` -- prime number > 0

    - ``k`` -- integer > 0

    - ``m`` -- integer (depending only on mod `p^k`)

    - ``zvec``, ``nzvec`` -- lists of integers up to dim(`Q`)

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,2,3])
        sage: Q.count_congruence_solutions__bad_type_II(3, 1, 0, None, None)
        2
    """
    return CountAllLocalTypesNaive(self, p, k, m, zvec, nzvec)[5]
